import csv
from io import TextIOBase
import json
from datetime import datetime
from typing import Dict, List, Union

import xmltodict
from rb.cna.cna_graph import CnaGraph
from rb.core.cscl.community import Community
from rb.core.cscl.contribution import Contribution
from rb.core.cscl.conversation import Conversation
from rb.core.cscl.cna_indices_enum import CNAIndices
from rb.core.cscl.participant import Participant
from rb.core.lang import Lang
from rb.core.text_element import TextElement
from rb.core.text_element_type import TextElementType
from rb.similarity.vector_model import VectorModelType
from rb.similarity.vector_model_factory import create_vector_model
from rb.utils.rblogger import Logger
from rb.processings.diacritics.DiacriticsRestoration import DiacriticsRestoration


logger = Logger.get_logger()

CONTRIBUTIONS_KEY = 'contributions'
ID_KEY = 'id'
PARENT_ID_KEY = 'parent_id'
TIMESTAMP_KEY = 'timestamp'
TEXT_KEY = 'text'
USER_KEY = 'user'
CONV_ID = 'conv_id'

JSONS_PATH = './jsons/'

FORMATS = [
    "%Y-%m-%d %H:%M:%S.%f %Z", 
    "%Y-%m-%d %H:%M:%S %Z", 
    "%Y-%m-%d %H:%M %Z",
    "%Y-%m-%d %H:%M:%S.%f", 
    "%Y-%m-%d %H:%M:%S", 
    "%Y-%m-%d %H:%M",
    "%d/%m/%Y %H.%M.%S.%f",
    "%d/%m/%Y %H:%M:%S.%f",
    "%H.%M.%S",
    "%H:%M:%S",
]

def get_json_from_json_file(filename: str) -> Dict:
    conversation_thread = dict()
    contribution_list = []

    with open(filename, "rt", encoding='utf-8') as json_file:
        contribution_list_json = json.load(json_file)
        for cnt in contribution_list_json:
            contribution = {
                ID_KEY: cnt['genid'],
                PARENT_ID_KEY: cnt["ref"],
                TIMESTAMP_KEY: int(float(cnt['time'])),
                USER_KEY: cnt["nickname"],
                TEXT_KEY: cnt["text"],
            }
            contribution_list.append(contribution)

    conversation_thread[CONTRIBUTIONS_KEY] = contribution_list

    return conversation_thread

def get_json_from_csv(filename: str) -> Dict:
    conversation_thread = dict()
    contribution_list = []

    with open(filename) as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=',')
        first = True			

        for row in csv_reader:
            if first:
                first = False
            else:
                contribution = dict()

                contribution[ID_KEY] = row[0]
                contribution[PARENT_ID_KEY] = row[1]
                contribution[USER_KEY] = row[2]
                contribution[TEXT_KEY] = row[3]
                contribution[TIMESTAMP_KEY] = row[4]

                contribution_list.append(contribution)

    conversation_thread[CONTRIBUTIONS_KEY] = contribution_list

    return conversation_thread

def parse_large_csv(filename: str) -> Dict:
    conversation_thread = dict()
    contribution_list = []

    with open(filename) as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=',')
        first = True			

        for row in csv_reader:
            if first:
                first = False
            else:
                contribution = dict()

                contribution[ID_KEY] = row[0]
                contribution[PARENT_ID_KEY] = row[1]
                if row[1] == '':
                    contribution[PARENT_ID_KEY] = '-1'

                contribution[USER_KEY] = row[2]
                contribution[TEXT_KEY] = row[6]

                contribution[TIMESTAMP_KEY] = read_date(row[3])

                contribution_list.append(contribution)

    conversation_thread[CONTRIBUTIONS_KEY] = contribution_list

    return conversation_thread

def read_date(date) -> datetime:
    for date_format in FORMATS:
        try:
            return datetime.strptime(date, date_format)
        except:
            pass
    

def load_from_xml(filename: Union[str, TextIOBase], diacritics_model: DiacriticsRestoration = None) -> Dict:
    if isinstance(filename, str):
        with open(filename, "rt", encoding="utf-8") as f:
            content = f.read()
    else:
        content = filename.read()
    my_dict = xmltodict.parse(content)
    if "corpus" in my_dict:
        my_dict = my_dict["corpus"]
    turns = my_dict["Dialog"]["Body"]["Turn"]
    if not isinstance(turns, List):
        turns = [turns]
    contributions = [
        {
            ID_KEY: int(utterance["@genid"]) - 1,
            PARENT_ID_KEY: int(utterance["@ref"]) - 1 if utterance["@ref"] else -1,
            TIMESTAMP_KEY: read_date(utterance["@time"]),
            USER_KEY: turn["@nickname"],
            TEXT_KEY: diacritics_model.process_string(utterance["#text"], mode="replace_missing") if diacritics_model else utterance["#text"],
        }
        for turn in turns
        for utterance in (turn["Utterance"] if isinstance(turn["Utterance"], List) else [turn["Utterance"]])
        if "#text" in utterance
    ]

    return {CONTRIBUTIONS_KEY: contributions, CONV_ID: my_dict["Dialog"]["@team"]}


def export_individual_statistics(participants: List[Participant], filename: str):
	first = True

	cscl_keys = [CNAIndices.CONTRIBUTIONS_SCORE, CNAIndices.SOCIAL_KB, CNAIndices.NO_CONTRIBUTION, CNAIndices.OUTDEGREE, CNAIndices.INDEGREE,
				CNAIndices.NO_NEW_THREADS, CNAIndices.NEW_THREADS_OVERALL_SCORE, CNAIndices.NEW_THREADS_CUMULATIVE_SOCIAL_KB,
				CNAIndices.AVERAGE_LENGTH_NEW_THREADS]

	evaluate_interaction(conv)
	evaluate_involvement(conv)
	evaluate_textual_complexity(conv)
	perform_sna(conv, False)
	with open(filename, 'w') as f:
		for p in participants:
			indices = p.indices
			keys = ['participant'] + [str(k.value) for k in CNAIndices]
			values = [p.get_id()] + [str(p.get_index(k)) for k in cscl_keys]

			if first:
				f.write(','.join(keys) + '\n')
				first = False

			f.write(','.join(values) + '\n')
		f.flush()
		f.close()

def export_textual_complexity(participants: List[Participant], filename: str):
	first = True

	with open(filename, 'w') as f:
		for p in participants:
			indices = p.textual_complexity_indices
			keys = ['participant'] + [str(k) for k in indices.keys()]
			values = [p.get_id()] + [str(indices[k]) for k in indices.keys()]

			if first:
				f.write(','.join(keys) + '\n')
				first = False

			f.write(','.join(values) + '\n')
		f.flush()
		f.close()




from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import os
from typing import Dict, List, Tuple

import numpy as np
import tensorflow as tf
from collections import namedtuple

from rb.processings.gec.transformer.transformer import Transformer
from rb.processings.gec.transformer.transformer_scheduler import CustomSchedule
from rb.processings.gec.transformer.utils import create_masks
from rb.processings.gec import beam_search
from rb.utils.downloader import check_version, download_model
from rb.core.lang import Lang
import tensorflow_datasets as tfds



class Beam(namedtuple("Beam", ["log_prob", "ids", "sentence", "length"])):
  """A finished beam

  Args:
    probs: Log probability of them beam
    ids: List of ids of the the beam
    sentence: Sentence in string
    lengths: Length of the beam
  """
  pass

class GecCorrector():

    def __init__(self, d_model: str=64, beam: int=8,
                lm_path: str=None, normalize: bool=False, weight_lm: float=1., check_updates = True):
        self.d_model = d_model 
        self.beam = beam 
        self.lm_path = lm_path
        self.normalize = normalize
        self.weight_lm = weight_lm
        
        if d_model == 768:
            name = 'transformer_768'
        elif d_model == 64:
            name = 'transformer_64'
        
        self.checkpoint_path = os.path.join('resources', 'ro', 'models', 'gec', name)
        self.vocabulary_path = os.path.join(self.checkpoint_path, 'tokenizer_ro')
        if check_updates and check_version(Lang.RO, ['models', 'gec', name]):
            download_model(Lang.RO, ['models', 'gec', name])
        self.transformer, self.optimizer = self.__load_model()
        self.lm_model = self.__load_lm()
        self.tokenizer_ro = self.__load_tokenizer()
        
       

    def __load_tokenizer(self):
        tokenizer_ro = tfds.features.text.SubwordTextEncoder.load_from_file(self.vocabulary_path)
        return tokenizer_ro

    def __load_lm(self):
        if self.lm_path:
            # install kenlm from https://github.com/kpu/kenlm, but it works without it (no reranking for beams)
            import kenlm
            lm_model = kenlm.Model(lm_path)
        else:
            lm_model = None
        return lm_model
    
    def __load_model(self):
        
        self.__set_params_model(self.d_model)
        self.vocab_size = self.dict_size + 2

        learning_rate = CustomSchedule(self.d_model)
        optimizer = tf.keras.optimizers.Adam(learning_rate, beta_1=0.9, beta_2=0.98, 
                                        epsilon=1e-9)

        transformer = Transformer(self.num_layers, self.d_model, self.num_heads, self.dff,
                            self.vocab_size, self.vocab_size, 
                            pe_input=self.vocab_size, 
                            pe_target=self.vocab_size,
                            rate=self.dropout)

        ckpt = tf.train.Checkpoint(transformer=transformer, optimizer=optimizer)
        
        ckpt_manager = tf.train.CheckpointManager(ckpt, self.checkpoint_path, max_to_keep=5)
        if ckpt_manager.latest_checkpoint:
            ckpt.restore(ckpt_manager.latest_checkpoint)
        else:
            tf.compat.v1.logging.error('no checkpoints for transformers, aborting')
            return None

        return transformer, optimizer

    def __set_params_model(self, d_model: int=64):
        if d_model == 64:
            self.num_layers = 2
            self.seq_length = 64
            self.dff = 64
            self.num_heads = 2
            self.dropout = 0.1
            self.dict_size = 1024

            self.num_layers = 2
        elif d_model == 768:
            self.num_layers = 6
            self.seq_length = 512
            self.dff = 2048
            self.num_heads = 8
            self.dropout = 0.1
            self.dict_size = (2**15)

    def __init_beam(self, vocab_size, end_token_id, beam_width=1):
    
        length_penalty = 0.6 if self.normalize else 0.0
        config = beam_search.BeamSearchConfig(
            beam_width=beam_width,
            vocab_size=vocab_size,
            eos_token=end_token_id,
            length_penalty_weight=length_penalty,
            choose_successors_fn=beam_search.choose_top_k)

        beam_state = beam_search.BeamSearchState(
            log_probs=tf.nn.log_softmax(tf.ones(config.beam_width)),
            lengths=tf.constant(
                1, shape=[config.beam_width], dtype=tf.int32),
            finished=tf.zeros(
                [config.beam_width], dtype=tf.bool))
        return config, beam_state

    def correct_sentence(self, inp_sentence: str): 
        start_token, end_token = [self.tokenizer_ro.vocab_size], [self.tokenizer_ro.vocab_size + 1]
        inp_sentence = inp_sentence.strip()
        in_sentence = inp_sentence
        inp_sentence = start_token + self.tokenizer_ro.encode(inp_sentence) + end_token
        if len(inp_sentence) > self.d_model:
            print('Sentence is too long (more than 766 subwords), cannot correct it.')
            return in_sentence

        start_token_id, end_token_id = self.tokenizer_ro.vocab_size, self.tokenizer_ro.vocab_size + 1

        # duplicate x beam_width == batch size
        encoder_input = tf.expand_dims(inp_sentence, 0)
        encoder_input = tf.tile(encoder_input, [self.beam, 1])

        decoder_input = [start_token_id] * self.beam
        output = tf.expand_dims(decoder_input, 1) # for batch size == beam_wisth

        # beam search init 
        config, beam_state = self.__init_beam(vocab_size=(self.dict_size + 2),
                                                    end_token_id=end_token_id, 
                                                    beam_width=self.beam)
        beam_values = tf.constant(start_token_id, shape=(1, self.beam))
        beam_parents = tf.zeros((2, self.beam), dtype=tf.int32)
        
        for i in range(self.d_model + 2):
            enc_padding_mask, combined_mask, dec_padding_mask = create_masks(
                encoder_input, output)

            predictions, attention_weights = self.transformer(encoder_input, output,
                                                                False, enc_padding_mask,
                                                                combined_mask, dec_padding_mask)
            # !predictions.shape == (batch_size, i, vocab_size) (predicts a softmax for each existing word)
            beam_pred = tf.squeeze(predictions[: ,-1:, :], 1)  # (batch_size, 1, vocab_size), select only the last word
            bs_output, beam_state = beam_search.beam_search_step(time_=i, logits=beam_pred,
                                                                beam_state=beam_state, config=config)

            # add new predictions to the beams decoder
            bs_output_predicted_ids = tf.expand_dims(bs_output.predicted_ids, axis=0)
            beam_values = tf.concat([beam_values, bs_output_predicted_ids], axis=0)
            res = tf.cast(beam_search.gather_tree_py(beam_values.numpy(), beam_parents.numpy()), dtype=tf.int32)
            output = tf.transpose(res)

            bs_output_beam_parent_ids = tf.expand_dims(bs_output.beam_parent_ids, axis=0)
            beam_parents = tf.concat([beam_parents, bs_output_beam_parent_ids], axis=0)

            all_finished = tf.reduce_all(beam_state.finished) # and
            if all_finished:    break

        beams = []
        for i, out in enumerate(output):
            b = Beam(log_prob=beam_state.log_probs[i].numpy(), sentence="", ids=out.numpy(), length=len(out.numpy()))
            beams.append(b)

        sentences = self.__decode_sentence(beams)
        # for sent in sentences:
        #     print(sent.log_prob, sent.sentence)
        sentences = sorted(sentences, key=lambda beam: beam.log_prob, reverse=True)
        return sentences[0].sentence
    
    def __decode_sentence(self, beams):
        sentences = []

        for beam in beams:
            sentence_ids = []

            for i in beam.ids:
                if i < self.tokenizer_ro.vocab_size:
                    sentence_ids.append(i)
                if i == self.tokenizer_ro.vocab_size + 1:
                    break
                
            predicted_sentence = self.tokenizer_ro.decode(sentence_ids)

            if self.lm_model:
                lm_prob = lm_model.score(predicted_sentence, bos=True, eos=True)

                if self.normalize:
                    cand_prob = beam.log_prob + self.weight_lm * lm_prob * (1.0/beam.length)
                else:
                    cand_prob = beam.log_prob + self.weight_lm * lm_prob
            else:
                cand_prob = beam.log_prob

            sentences.append(Beam(log_prob=cand_prob, sentence=predicted_sentence,
                                 ids=beam.ids, length=len(beam.ids)))
        return sentences


if __name__ == "__main__":
    corrector = GecCorrector(checkpoint_path='checkpoints/10m_transformer_768_2',
                vocabulary_path='checkpoints/10m_transformer_768_2/tokenizer_ro', d_model=768)

    decoded_senntence = corrector.correct_sentence('Am mers la magazi sa cuumper leegume si fructe de mancaat.')
    print(decoded_senntence)


from typing import Dict, List
from munkres import Munkres, DISALLOWED
from rb.docs_processing.author import Author
from rb.docs_processing.article import Article
from rb.similarity.vector_model import VectorModel
from rb.similarity.transformers_encoder import TransformersEncoder

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np


class Graph:
    EDGE_THRESHOLD = 0.064
    BIG_VALUE = 10000
    NO_EDGE_PENALTY = 100
    # EDGE_THRESHOLD = 1.01

    def __init__(self, semantic_models: List[VectorModel]):
        self.authors_dict: Dict[str, Author] = {}
        self.articles_set = set()
        self.authors_set = set()
        self.adjacent_list = {}
        self.authors_ranking_graph = {}
        self.articles_distances_dict = {}
        self.semantic_models = semantic_models
        self.articles_mean = 0.0
        self.articles_std = 0.0
        self.authors_mean = 0.0
        self.authors_std = 0.0

    def extract_authors_dict(self) -> None:
        self.authors_set = set([x for _, x in self.authors_dict.items()])

    def add_to_adjacent_list(self, entity1, entity2, dist: float, link_type: str, direction: str = "bidirectional") -> None:
        if entity1 in self.adjacent_list:
            self.adjacent_list[entity1].append((entity2, dist, link_type))
        else:
            self.adjacent_list[entity1] = [(entity2, dist, link_type)]

        if direction == "bidirectional":
            if entity2 in self.adjacent_list:
                self.adjacent_list[entity2].append((entity1, dist, link_type))
            else:
                self.adjacent_list[entity2] = [(entity1, dist, link_type)]

    def build_edges_and_construct_author_rankings_graph(self, authors_relation_type="bidirectional") -> None:
        articles_list = list(self.articles_set)
        authors_list = list(self.authors_set)

        self.build_edges_between_articles_and_save_the_distances(articles_list)
        if not authors_relation_type:
            return
        
        self.build_edges_between_articles_and_authors(articles_list, authors_list)
        authors_semantic_similarities_dictionary = self.compute_semantic_similarities_between_authors(authors_list, authors_relation_type)
        self.build_authors_ranking_graph(authors_semantic_similarities_dictionary, authors_relation_type=authors_relation_type)
        self.build_edges_between_authors(authors_semantic_similarities_dictionary, authors_relation_type=authors_relation_type)

    def build_edges_between_articles_and_authors(self, articles_list, authors_list):
        for i in range(len(articles_list)):
            for j in range(len(authors_list)):
                dist = compute_semantic_distance_between_author_and_article(authors_list[j], articles_list[i],
                                                                            self.semantic_models)
                if dist < Graph.EDGE_THRESHOLD:
                    self.add_to_adjacent_list(authors_list[j], articles_list[i], dist, 'aut-art')

    def build_edges_between_articles_and_save_the_distances(self, articles_list):
        counter = 0
        articles_to_distances_dict = self.build_article_pairs_to_distance_dict(articles_list)
        threshold, self.articles_mean, self.articles_std = \
            compute_edge_threshold_mean_and_std(articles_to_distances_dict)

        for articles_pair, distance in articles_to_distances_dict.items():
            article1 = articles_pair[0]
            article2 = articles_pair[1]
            if distance < threshold:
                counter += 1
                # print(counter)
                self.add_to_adjacent_list(article1, article2, distance, 'art-art')
                self.articles_distances_dict[(article1, article2)] = distance
            else:
                self.articles_distances_dict[(article1, article2)] = 1.0

    def build_article_pairs_to_distance_dict(self, articles_list):
        articles_to_distances_dict = {}
        for i in range(len(articles_list) - 1):
            for j in range(i + 1, len(articles_list)):
                article1 = articles_list[i]
                article2 = articles_list[j]
                distance = compute_distance_between_articles(article1, article2, self.semantic_models)
                articles_to_distances_dict[(article1, article2)] = distance
        return articles_to_distances_dict

    def build_edges_between_authors(self, authors_semantic_similarities_dictionary, authors_relation_type):
        if authors_relation_type == "bidirectional":
            similarities = authors_semantic_similarities_dictionary.values()
            maximum_value, minimal_value, self.authors_mean, self.authors_std = extract_min_max_mean_std(similarities)
            threshold = self.authors_mean - self.authors_std
            for authors_pair, similarity in authors_semantic_similarities_dictionary.items():
                scaled_distance = 1 - (similarity - minimal_value) / (maximum_value - minimal_value)
                if scaled_distance < threshold:
                    self.add_to_adjacent_list(authors_pair[0], authors_pair[1], scaled_distance, 'aut-aut')
        elif authors_relation_type == 'unidirectional':
            distances = [1 - x for x in authors_semantic_similarities_dictionary.values()]
            self.authors_mean = np.mean(distances)
            self.authors_std = np.std(distances)
            for authors_pair, similarity in authors_semantic_similarities_dictionary.items():
                self.add_to_adjacent_list(authors_pair[0], authors_pair[1], 1 - similarity, 'aut-aut', direction='unidirectional')
        else:
            raise Exception("Unknown authors relation type")

    def build_authors_ranking_graph(self, authors_semantic_similarities_dictionary, authors_relation_type='bidirectional'):
        for authors_pair, similarity in authors_semantic_similarities_dictionary.items():
            author1 = authors_pair[0]
            author2 = authors_pair[1]
            if authors_relation_type == 'bidirectional':
                self.authors_ranking_graph[author1] = self.authors_ranking_graph.get(author1, [])
                self.authors_ranking_graph[author1].append((author2, similarity))
                self.authors_ranking_graph[author2] = self.authors_ranking_graph.get(author2, [])
                self.authors_ranking_graph[author2].append((author1, similarity))
            elif authors_relation_type == 'unidirectional':
                self.authors_ranking_graph[author1] = self.authors_ranking_graph.get(author1, [])
                self.authors_ranking_graph[author1].append((author2, similarity))
            else:
                raise Exception('Invalid authors_relation_type')

    def compute_semantic_similarities_between_authors(self, authors_list, authors_relation_type="bidirectional"):
        authors_semantic_similarities_dictionary = {}
        for i in range(len(authors_list) - 1):
            for j in range(i + 1, len(authors_list)):
                author1 = authors_list[i]
                author2 = authors_list[j]
                if authors_relation_type == "bidirectional":
                    similarity = self.compute_semantic_similarity_between_authors(author1, author2)
                    authors_semantic_similarities_dictionary[(author1, author2)] = similarity
                elif authors_relation_type == "unidirectional":
                    print(f"Computing semantic similarity between {author1} with {len(author1.articles)} and {author2} with {len(author2.articles)}")
                    similarity = self.compute_semantic_similarity_between_author1_and_author2(author1, author2)
                    authors_semantic_similarities_dictionary[(author1, author2)] = similarity
                    similarity = self.compute_semantic_similarity_between_author1_and_author2(author2, author1)
                    authors_semantic_similarities_dictionary[(author2, author1)] = similarity
                else:
                    raise Exception("Invalid authors_relation_type")
                # if dist < max_distance:
                #     self.add_to_adjacent_list(authors_list[i], authors_list[j], dist, 'aut-aut')
        return authors_semantic_similarities_dictionary

    def get_authors_by_type_degree(self, max_=None):
        authors = [(author, sum(n for _, n in self.authors_ranking_graph[author])) for author in self.authors_set]
        if max_:
            return sorted(authors, key=lambda x: x[1], reverse=True)[:max_]
        return sorted(authors, key=lambda x: x[1], reverse=True)
    
    
    def get_authors_by_out_type_degree(self, max_=None):
        return self.get_authors_by_type_degree(max_)
    
    
    def get_authors_by_in_type_degree(self, max_=None):
        inverse_dict = {}
        for author in self.authors_set:
            for neighbour, similarity in self.authors_ranking_graph[author]:
                inverse_dict[neighbour] = inverse_dict.get(neighbour, 0) + similarity
        authors = [(author, inverse_dict[author]) for author in self.authors_set]
        if max_:
            return sorted(authors, key=lambda x: x[1], reverse=True)[:max_]
        return sorted(authors, key=lambda x: x[1], reverse=True)
    

    def get_articles_by_type_degree(self, max_=None):
        articles = [(article, sum(1 - n for _, n, t in self.adjacent_list[article] if t == 'art-art'),
                     len([t for _, _, t in self.adjacent_list[article] if t == 'art-art'])) for article in
                    self.articles_set if article in self.adjacent_list]
        if max_:
            return sorted(articles, key=lambda x: x[1], reverse=True)[:max_]
        return sorted(articles, key=lambda x: x[1], reverse=True)

    def get_similarities_between_articles(self):
        similarities_dictionary = {}
        for pair in self.get_distances_between_articles():
            similarities_dictionary[(pair[0], pair[1])] = 1 - pair[2]
        return similarities_dictionary

    def get_similarities_between_authors(self):
        similarities_dictionary = {}
        for author, neighbours in self.authors_ranking_graph.items():
            for neighbour in neighbours:
                if (neighbour[0], author) not in similarities_dictionary:
                    similarities_dictionary[(author, neighbour[0])] = neighbour[1]
        return similarities_dictionary

    def get_distances_between_articles(self):
        return self.get_distances_between_nodes_of_same_type("art-art", self.articles_set)

    def get_distances_between_authors(self):
        return self.get_distances_between_nodes_of_same_type("aut-aut", self.authors_set)

    def get_distances_between_nodes_of_same_type(self, node_type, node_set):
        pairs = []
        used_nodes = set()
        for node in node_set:
            if node in self.adjacent_list:
                for neigh, score, edge_type in self.adjacent_list[node]:
                    if edge_type == node_type and neigh not in used_nodes:
                        pairs.append((node, neigh, score))
                used_nodes.add(node)
        return pairs

    def plot_histogram_of_articles_distance(self):
        distances = [pair[2] for pair in self.get_distances_between_articles()]
        print(np.mean(distances))
        print(np.std(distances))
        plt.hist(distances, color="blue", edgecolor="black")
        plt.show()
        
    # only one direction, author1 -> author2
    def compute_semantic_similarity_between_author1_and_author2(self, author1, author2):
        if author1 == author2:
            return len(author1.articles)
        
        if len(author1.articles) >= len(author2.articles):
            return self.compute_semantic_similarity_between_authors(author1, author2)
        
        to_be_excluded_author2_articles = set()
        similarities = []
        while True:
            distance_dict = self.build_articles_distance_dictionary(author1, author2, to_be_excluded_author2_articles)
            if len(distance_dict) == 1:
                similarities.append(1 - next(iter(distance_dict.values())))
                return np.mean(similarities)
            
            if not distance_dict:
                return np.mean(similarities)
            
            author1_name_articles_tuple = (author1.name, list(set([x[0] for x in distance_dict.keys()])))
            author2_name_articles_tuple = (author2.name, list(set([x[1] for x in distance_dict.keys()])))
            
            articles_distances_matrix = build_distances_between_articles_matrix(
                author1_name_articles_tuple[1],
                author2_name_articles_tuple[1],
                distance_dict)
            
            # check if no edges
            are_edges = False
            for e in articles_distances_matrix:
                for i in e:
                    if i != Graph.BIG_VALUE:
                        are_edges = True
                        break
                if are_edges:
                    break
            if not are_edges:
                similarities += [0] * min(len(author1.articles), len(author2.articles) - len(to_be_excluded_author2_articles))
                return np.mean(similarities)
            
            munkres_client = Munkres()
            try:
                indexes = munkres_client.compute(articles_distances_matrix)
            except:
                if similarities:
                    return max(np.mean(similarities) - 0.001, 0) # epsilon in case of an error
                return 0.0  # minimum similarity possible
            
            for row, column in indexes:
                if articles_distances_matrix[row][column] == Graph.BIG_VALUE:
                    continue
                value = 1 - articles_distances_matrix[row][column] / 100
                similarities.append(value)
                to_be_excluded_author2_articles.add(author2_name_articles_tuple[1][column])


    def compute_semantic_similarity_between_authors(self, author1, author2):
        if author1 == author2:
            return len(author1.articles)

        distance_dict = self.build_articles_distance_dictionary(author1, author2)
        # (distance_dict)
        if len(distance_dict) == 1:
            return 1 - next(iter(distance_dict.values()))

        if distance_dict:
            a1 = (author1.name, list(set([x[0] for x in distance_dict.keys()])))
            a2 = (author2.name, list(set([x[1] for x in distance_dict.keys()])))
            return compute_maximum_coupling_of_maximum_similarity(a1, a2, distance_dict)

        return 0.0
    

    def build_articles_distance_dictionary(self, author1, author2, to_be_excluded_author2_articles=set()):
        distance_dict = {}
        for article1 in author1.articles:
            for article2 in author2.articles:
                if article2.title in to_be_excluded_author2_articles:
                    continue
                if article1 == article2:
                    distance_dict[(article1.title, article2.title)] = 1 - (1 / len(article1.author_list))
                else:
                    distance_dict[(article1.title, article2.title)] = \
                        self.articles_distances_dict.get((article1, article2), 0) + \
                        self.articles_distances_dict.get((article2, article1), 0)
        return distance_dict


def compute_maximum_coupling_of_maximum_similarity(author1_name_articles_tuple, author2_name_articles_tuple,
                                                   distance_dictionary):
    articles_distances_matrix = build_distances_between_articles_matrix(
        author1_name_articles_tuple[1],
        author2_name_articles_tuple[1],
        distance_dictionary)
    # print(articles_distances_matrix)
    munkres_client = Munkres()
    try:
        indexes = munkres_client.compute(articles_distances_matrix)
    except:
        return 0.0  # minimum similarity possible
    return compute_semantic_similarity_coupling_value(articles_distances_matrix, indexes)


def compute_semantic_similarity_coupling_value(articles_distances_matrix, indexes):
    total = 0
    for row, column in indexes:
        if articles_distances_matrix[row][column] == Graph.BIG_VALUE:
            continue
        value = 1 - articles_distances_matrix[row][column] / 100
        total += value
    return total


def clear_matrix_of_only_disallowed(distances_matrix, only_disallowed_lines_index):
    only_disallowed_lines_index.reverse()
    for index in only_disallowed_lines_index:
        del distances_matrix[index]

    if not distances_matrix:
        return

    only_disallowed_columns_index = []
    for j in range(len(distances_matrix[0])):
        was_only_disallowed = True
        for i in range(len(distances_matrix)):
            if distances_matrix[i][j] != DISALLOWED:
                was_only_disallowed = False
                break
        if was_only_disallowed:
            only_disallowed_columns_index.append(j)

    only_disallowed_columns_index.reverse()

    for line in distances_matrix:
        for index in only_disallowed_columns_index:
            del line[index]


def build_distances_between_articles_matrix(author1_articles, author2_articles,
                                            distance_dictionary):
    distances_matrix = [[Graph.BIG_VALUE for x in range(len(author2_articles))] for y in
                        range(len(author1_articles))]
    for i1, art1 in enumerate(author1_articles):
        for i2, art2 in enumerate(author2_articles):
            if (art1, art2) in distance_dictionary and distance_dictionary[(art1, art2)] < Graph.EDGE_THRESHOLD:
                distances_matrix[i1][i2] = distance_dictionary[(art1, art2)] * 100
            elif (art2, art1) in distance_dictionary and distance_dictionary[(art2, art1)] < Graph.EDGE_THRESHOLD:
                distances_matrix[i1][i2] = distance_dictionary[(art2, art1)] * 100

    return distances_matrix


def compute_distance_between_articles(article1: Article, article2: Article,
                                      semantic_models: List[VectorModel]) -> float:
    if article1 == article2:
        return 0.0

    similarity = 0.0
    for model in semantic_models:
        if type(model) is TransformersEncoder:
            similarity += model.similarity(article1.document.vectors[model], article2.document.vectors[model])
        else:
            similarity += model.similarity(article1.document, article2.document)

    if similarity != 0:
        distance = 1 - similarity / len(semantic_models)
        return distance if distance >= 0 else 0
    return 1.0


def compute_semantic_distance_between_author_and_article(author: Author, article: Article,
                                                         semantic_models: List[VectorModel]) -> float:
    if article in author.articles:
        return 0.0

    distance = 0.0
    for aa in author.articles:
        distance += compute_distance_between_articles(aa, article, semantic_models)

    if distance > 0:
        return distance / len(author.articles)

    return 1.0


def extract_min_max_mean_std(similarities):
    minimal_value = min(similarities)
    maximum_value = max(similarities)
    scaled_distances = [1 - (similarity - minimal_value) / (maximum_value - minimal_value)
                        for similarity in similarities]
    mean = np.mean(scaled_distances)
    std = np.std(scaled_distances)
    return maximum_value, minimal_value, mean, std


def compute_edge_threshold_mean_and_std(articles_to_distances_dict):
    values = articles_to_distances_dict.values()
    mean = np.mean(list(values))
    std = np.std(list(values))
    # print(f"mean: {mean}, std: {std}")
    threshold = mean - std
    return threshold, mean, std



from typing import Dict, List, Union

import networkx as nx
from networkx.algorithms.centrality import betweenness_centrality, closeness_centrality, eigenvector_centrality
from rb.cna.cna_graph import CnaGraph
from rb.complexity.complexity_index import compute_indices
from rb.core.block import Block
from rb.core.cscl.cna_indices_enum import CNAIndices
from rb.core.cscl.contribution import Contribution
from rb.core.cscl.conversation import Conversation
from rb.core.cscl.community import Community
from rb.core.lang import Lang
from rb.core.pos import POS
from rb.core.text_element import TextElement
from rb.utils.rblogger import Logger


def get_block_importance(graph: nx.DiGraph, a: Block, b: Block) -> float:
    return graph.edges[a, b]["weight"] if (a, b) in graph.edges else 0.


def evaluate_interaction(conversation: Union[Conversation, Community]):
    cna_graph = conversation.container.graph if conversation.container is not None else conversation.graph
    
    participants = conversation.get_participants()
    if isinstance(conversation, Community):
        contributions = [contribution for conv in conversation.components for contribution in conv.get_contributions()]
    else:
        contributions = conversation.get_contributions()
    conversation.init_scores()

    if (len(participants) == 0):
        return

    for i, contribution1 in enumerate(contributions):
        p1 = contribution1.get_participant().get_id()

        for contribution2 in contributions[:i]:
            weight = get_block_importance(cna_graph.filtered_graph, contribution1, contribution2)
            if weight > 0:
                p2 = contribution2.get_participant().get_id()
                conversation.update_score(p1, p2, weight)


def evaluate_involvement(conversation: Union[Conversation, Community]):
    cna_graph = conversation.container.graph if conversation.container is not None else conversation.graph
    participants = conversation.get_participants()

    if (len(participants) == 0):
        return
    if isinstance(conversation, Community):
        conversations = conversation.components
    else:
        conversations = [conversation]
    for conv in conversations:
        for i, contribution in enumerate(conv.get_contributions()):
            p = conversation.participant_map[contribution.get_participant().get_id()]

            current_value = p.get_index(CNAIndices.CONTRIBUTIONS_SCORE)
            p.set_index(CNAIndices.CONTRIBUTIONS_SCORE, current_value + cna_graph.importance[contribution])
            original_conv = conversation.container.get_conversation(conv.conv_id) if conversation.container else conv
            added_value = 0
            for prev in original_conv.get_contributions():
                if contribution.index == prev.index:
                    break
                if p != prev.get_participant():
                    added_value += get_block_importance(cna_graph.filtered_graph, contribution, prev)
            p.set_index(CNAIndices.SOCIAL_KB, p.get_index(CNAIndices.SOCIAL_KB) + added_value)

def evaluate_textual_complexity(conversation: Conversation):
    participants = conversation.get_participants()
    
    for p in participants:
        contributions = conversation.get_participant_contributions(p.get_id())
        p.set_eligible_contributions(contributions)

        Conversation.create_participant_conversation(conversation.lang, p)

        own_conversation = p.own_conversation

        print('Begin computing textual complexity')
        compute_indices(doc=own_conversation)
        print('Finished computing textual complexity')

        for key, value in own_conversation.indices.items():
            p.set_textual_index(repr(key), value)


def perform_sna(conversation: Union[Conversation, Community], needs_anonymization = False) -> nx.DiGraph:
    participants = conversation.get_participants()
    cna_graph = conversation.graph
    participants_graph = nx.DiGraph()
    participants_graph.add_nodes_from(participants)
    conversations = conversation.get_conversations() if isinstance(conversation, Community) else [conversation]
    for conv in conversations:
        for i, u in enumerate(conv.get_contributions()):
            for v in conv.get_contributions()[(i+1):]:
                if u.get_participant() != v.get_participant() and u.timestamp < v.timestamp:
                    edge = cna_graph.filtered_graph.get_edge_data(u, v)
                    if edge is None:
                        continue
                    weight = edge["weight"]
                    if participants_graph.has_edge(u.get_participant(), v.get_participant()):
                        participants_graph[u.get_participant()][v.get_participant()]["weight"] += weight
                    else:
                        participants_graph.add_edge(u.get_participant(), v.get_participant(), weight=weight)
    betweeness = betweenness_centrality(participants_graph)
    closeness = closeness_centrality(participants_graph)
    eigen = eigenvector_centrality(participants_graph, max_iter=1000, tol=1.0e-5)
    for p in participants:
        p.set_index(CNAIndices.OUTDEGREE, participants_graph.out_degree(p, weight="weight"))
        p.set_index(CNAIndices.INDEGREE, participants_graph.in_degree(p, weight="weight"))
        p.set_index(CNAIndices.BETWEENNESS, betweeness[p])
        p.set_index(CNAIndices.CLOSENESS, closeness[p])
        p.set_index(CNAIndices.EIGENVECTOR, eigen[p])
    return participants_graph



#!/usr/bin/env python
# coding: utf8
"""Visualize spaCy word vectors in Tensorboard.

Adapted from: https://gist.github.com/BrikerMan/7bd4e4bd0a00ac9076986148afc06507
"""
from __future__ import unicode_literals

from os import path

import math
import numpy
import spacy
import tensorflow as tf
from tensorflow.contrib.tensorboard.plugins.projector import visualize_embeddings, ProjectorConfig
from gensim.models import LsiModel 


def load_model(inputFile: str):
    vocab = []
    matrix = []
    with open(inputFile, "rt") as f:
        n, dim = f.readline().split(" ")
        for line in f.readlines():
            cols = line.split(" ")
            word = cols[0]
            vector = [float(val) for val in cols[1:]]
            vocab.append(word)
            matrix.append(vector)
    return vocab, matrix

if __name__ == "__main__":
    out_loc = "models"
    model_path = "resources/new_config/semantic-models-enea_tasa-en/word2vec.model"
    meta_file = "{}.tsv".format("vocab")
    out_meta_file = path.join(out_loc, meta_file)
    name = "enea-word2vec"
    
    print('Loading vectors model: {}'.format(model_path))
    vocab, matrix = load_model(model_path)

    # model = LsiModel.load("RO/ReadME/lsa.bin")
    # matrix = matrix * numpy.sqrt(model.projection.s)
    
    print('Building Tensorboard Projector metadata for ({}) vectors: {}'.format(len(vocab), meta_file))

    # Store vector data in a tensorflow variable
    
    # Write a tab-separated file that contains information about the vectors for visualization
    #
    # Reference: https://www.tensorflow.org/programmers_guide/embedding#metadata
    with open(out_meta_file, 'wb') as file_metadata:
        # Define columns in the first row
        # file_metadata.write("Text\n".encode('utf-8'))
        # Write out a row for each vector that we add to the tensorflow variable we created
        for word in vocab:
            # https://github.com/tensorflow/tensorflow/issues/9094
            
            # Store vector data and metadata
            file_metadata.write("{}\n".format(word).encode('utf-8'))
            
    print('Running Tensorflow Session...')
    sess = tf.InteractiveSession()
    tf.Variable(matrix, trainable=False, name=name)
    tf.global_variables_initializer().run()
    saver = tf.train.Saver()
    writer = tf.summary.FileWriter(out_loc, sess.graph)

    # Link the embeddings into the config
    config = ProjectorConfig()
    embed = config.embeddings.add()
    embed.tensor_name = name
    embed.metadata_path = meta_file

    # Tell the projector about the configured embeddings and metadata file
    visualize_embeddings(writer, config)

    # Save session and print run command to the output
    print('Saving Tensorboard Session...')
    saver.save(sess, path.join(out_loc, '{}.ckpt'.format(name)))
    print('Done. Run `tensorboard --logdir={0}` to view in Tensorboard'.format(out_loc))




import tensorflow as tf

class CustomSchedule(tf.keras.optimizers.schedules.LearningRateSchedule):
    def __init__(self, d_model, warmup_steps=4000):
        super(CustomSchedule, self).__init__()
        
        self.d_model = d_model
        self.d_model = tf.cast(self.d_model, tf.float32)

        self.warmup_steps = warmup_steps
        
    def __call__(self, step):
        arg1 = tf.math.rsqrt(step)
        arg2 = step * (self.warmup_steps ** -1.5)
        
        return tf.math.rsqrt(self.d_model) * tf.math.minimum(arg1, arg2)


from enum import Enum, auto

class ConnTypeEnum(Enum):
    COORD = auto()
    LOGIC = auto()
    SEMI = auto()
    QUASI = auto()
    CONJ = auto()
    DISJ = auto()
    SIMPLE = auto()
    COMPLEX = auto()
    COORDCONJ = auto()
    ADDITION = auto()
    CONTRAST = auto()
    LINK = auto()
    ORDER = auto()
    REFERENCE = auto()
    REASONANDPURPOSE = auto()
    COND = auto()
    CONCC = auto()
    OPPOSIT = auto()
    TEMPORAL = auto()
    CONJUNCTS = auto()
    




class Author:

    def __init__(self, name: str):
        self.name = name
        self.articles = set()

    def __eq__(self, other):
        if isinstance(other, Author):
            return self.name == other.name
        return NotImplemented

    def __hash__(self):
        return hash(tuple(self.name))
    
    def __str__(self):
        return self.name
    
    def __repr__(self) -> str:
        return self.name


from typing import Dict, List

from rb.cna.cna_graph import CnaGraph
from rb.complexity.complexity_index import compute_indices
from rb.core.block import Block
from rb.core.cscl.cna_indices_enum import CNAIndices
from rb.core.cscl.community import Community
from rb.core.cscl.contribution import Contribution
from rb.core.cscl.conversation import Conversation
from rb.processings.cscl.participant_evaluation import get_block_importance, perform_sna
from rb.utils.rblogger import Logger

def compute_new_threads_metrics(community: Community):
    cna_graph = community.graph
    importance = cna_graph.importance
    
    for conversation in community.get_conversations():
        participant = conversation.get_contributions()[0].get_participant()
        p = community.get_participant(participant.get_id())
        current_value = p.get_index(CNAIndices.NO_NEW_THREADS)
        p.set_index(CNAIndices.NO_NEW_THREADS, current_value + 1)
        for contribution in conversation.get_contributions():
            participant = contribution.get_participant()
            p = community.get_participant(participant.get_id())

            current_value = p.get_index(CNAIndices.NEW_THREADS_OVERALL_SCORE)
            p.set_index(CNAIndices.NEW_THREADS_OVERALL_SCORE,
                        current_value + importance[conversation])

            # current_value = p.get_index(
            #     CNAIndices.NEW_THREADS_CUMULATIVE_SOCIAL_KB)
            # TODO
            # total_kb = conversation.get_cumulative_social_kb()
            # p.set_index(CNAIndices.NEW_THREADS_CUMULATIVE_SOCIAL_KB,
            #             current_value + total_kb)

            # todo add voice pmi evolution
            current_value = p.get_index(CNAIndices.AVERAGE_LENGTH_NEW_THREADS)
            p.set_index(CNAIndices.AVERAGE_LENGTH_NEW_THREADS,
                        current_value + 1)
    for p in community.get_participants():
        new_threads = p.get_index(CNAIndices.NO_NEW_THREADS)
        total_length = p.get_index(CNAIndices.AVERAGE_LENGTH_NEW_THREADS)

        if new_threads > 0:
            average_length = total_length / new_threads
            p.set_index(CNAIndices.AVERAGE_LENGTH_NEW_THREADS, average_length)


import numpy as np
import tensorflow as tf

def scaled_dot_product_attention(q, k, v, mask):
    """Calculate the attention weights.
    q, k, v must have matching leading dimensions.
    k, v must have matching penultimate dimension, i.e.: seq_len_k = seq_len_v.
    The mask has different shapes depending on its type(padding or look ahead) 
    but it must be broadcastable for addition.
    
    Args:
        q: query shape == (..., seq_len_q, depth)
        k: key shape == (..., seq_len_k, depth)
        v: value shape == (..., seq_len_v, depth_v)
        mask: Float tensor with shape broadcastable 
            to (..., seq_len_q, seq_len_k). Defaults to None.
        
    Returns:
        output, attention_weights
    """

    matmul_qk = tf.matmul(q, k, transpose_b=True)  # (..., seq_len_q, seq_len_k)
    
    # scale matmul_qk
    dk = tf.cast(tf.shape(k)[-1], tf.float32)
    scaled_attention_logits = matmul_qk / tf.math.sqrt(dk)

    # add the mask to the scaled tensor.
    if mask is not None:
        scaled_attention_logits += (mask * -1e9)  

    # softmax is normalized on the last axis (seq_len_k) so that the scores
    # add up to 1.
    attention_weights = tf.nn.softmax(scaled_attention_logits, axis=-1)  # (..., seq_len_q, seq_len_k)

    output = tf.matmul(attention_weights, v)  # (..., seq_len_q, depth_v)

    return output, attention_weights

class MultiHeadAttention(tf.keras.layers.Layer):
    def __init__(self, d_model, num_heads):
        super(MultiHeadAttention, self).__init__()
        """ split q, k, v by nr of heads """ 
        self.num_heads = num_heads
        self.d_model = d_model

        assert d_model % self.num_heads == 0

        self.depth = d_model // self.num_heads

        self.wq = tf.keras.layers.Dense(d_model)
        self.wk = tf.keras.layers.Dense(d_model)
        self.wv = tf.keras.layers.Dense(d_model)

        self.dense = tf.keras.layers.Dense(d_model)
        
    def split_heads(self, x, batch_size):
        """Split the last dimension into (num_heads, depth).
        Transpose the result such that the shape is (batch_size, num_heads, seq_len, depth)
        """
        x = tf.reshape(x, (batch_size, -1, self.num_heads, self.depth))
        return tf.transpose(x, perm=[0, 2, 1, 3])

    def call(self, v, k, q, mask):
        batch_size = tf.shape(q)[0]

        q = self.wq(q)  # (batch_size, seq_len, d_model)
        k = self.wk(k)  # (batch_size, seq_len, d_model)
        v = self.wv(v)  # (batch_size, seq_len, d_model)

        q = self.split_heads(q, batch_size)  # (batch_size, num_heads, seq_len_q, depth)
        k = self.split_heads(k, batch_size)  # (batch_size, num_heads, seq_len_k, depth)
        v = self.split_heads(v, batch_size)  # (batch_size, num_heads, seq_len_v, depth)

        # scaled_attention.shape == (batch_size, num_heads, seq_len_q, depth)
        # attention_weights.shape == (batch_size, num_heads, seq_len_q, seq_len_k)
        scaled_attention, attention_weights = scaled_dot_product_attention(
            q, k, v, mask)

        scaled_attention = tf.transpose(scaled_attention, perm=[0, 2, 1, 3])  # (batch_size, seq_len_q, num_heads, depth)

        concat_attention = tf.reshape(scaled_attention, 
                                        (batch_size, -1, self.d_model))  # (batch_size, seq_len_q, d_model)

        output = self.dense(concat_attention)  # (batch_size, seq_len_q, d_model)
        
        return output, attention_weights


from rb.comprehension.utils.graph.cm_edge_do import CmEdgeDO
from rb.comprehension.utils.graph.cm_edge_type import CmEdgeType
from rb.comprehension.utils.graph.cm_node_do import CmNodeDO
from rb.comprehension.utils.graph.cm_node_type import CmNodeType

from rb.core.word import Word


class CmSyntacticGraph():

    def __init__(self):
        self.edge_list = []
        self.word_set = set()

    
    def index_edge(self, word1: Word, word2: Word) -> None:
        node1 = CmNodeDO(word1, CmNodeType.TextBased)
        node2 = CmNodeDO(word2, CmNodeType.TextBased)
        edge = CmEdgeDO(node1, node2, CmEdgeType.Syntactic, 1.0)

        if edge not in self.edge_list:
            self.edge_list.append(edge)
            self.word_set.add(word1)
            self.word_set.add(word2)



import tensorflow as tf
from rb.processings.gec.transformer.encoder import Encoder
from rb.processings.gec.transformer.decoder import Decoder

class Transformer(tf.keras.Model):
    def __init__(self, num_layers, d_model, num_heads, dff, input_vocab_size, 
                target_vocab_size, pe_input, pe_target, rate=0.1):
        super(Transformer, self).__init__()
        self.encoder = Encoder(num_layers, d_model, num_heads, dff, 
                            input_vocab_size, pe_input, rate)
        self.decoder = Decoder(num_layers, d_model, num_heads, dff, 
                            target_vocab_size, pe_target, rate)
        self.final_layer = tf.keras.layers.Dense(target_vocab_size)
        
    def call(self, inp, tar, training, enc_padding_mask, 
            look_ahead_mask, dec_padding_mask):
        enc_output = self.encoder(inp, training, enc_padding_mask)  # (batch_size, inp_seq_len, d_model)
        # dec_output.shape == (batch_size, tar_seq_len, d_model)

        dec_output, attention_weights = self.decoder(
            tar, enc_output, training, look_ahead_mask, dec_padding_mask)
        
        final_output = self.final_layer(dec_output)  # (batch_size, tar_seq_len, target_vocab_size)
        
        return final_output, attention_weights


from rb.similarity.vector_model import VectorModel
from rb.core.document import Document
from rb.core.lang import Lang
from rb.comprehension.utils.distance_strategies.cm_coref_indexer import CmCorefIndexer
from rb.comprehension.utils.distance_strategies.syntactic_word_distance_strategy import SyntacticWordDistanceStrategy
from rb.comprehension.cm_word_distance_indexer import CmWordDistanceIndexer
from typing import List
Models = List[VectorModel]


class CmIndexer():

    def __init__(self, text: str, lang: Lang, semantic_models: Models):
        self.text = text
        self.semantic_models = semantic_models
        self.lang = lang
        self.document = Document(lang, text)
        self.syntactic_indexer_list = []
        self.index_syntactic_distances()

    def index_syntactic_distances(self) -> None:
        coref_container = CmCorefIndexer(self.document, self.lang)
        sentences = self.document.get_sentences()
        sentence_number = 0
        for sentence in sentences:
            syntactic_graph = coref_container.get_cm_syntactic_graph(sentence, sentence_number)
            syntactic_strategy = SyntacticWordDistanceStrategy(syntactic_graph)
            wd_indexer = CmWordDistanceIndexer(list(syntactic_graph.word_set), syntactic_strategy)
            self.syntactic_indexer_list.append(wd_indexer)
            sentence_number += 1



from typing import Dict, List

from rb.core.lang import Lang
from rb.core.text_element import TextElement
from rb.core.text_element_type import TextElementType
from rb.utils.rblogger import Logger
import csv

class Participant:

    def __init__(self, participant_id: str):
        self.participant_id = participant_id
        self.indices = dict()
        self.longitudinal_indices = dict()
        self.textual_complexity_indices = dict()

        self.eligible_contributions = []
        self.significant_contributions = []

        self.own_conversation = None

    def get_id(self) -> str:
        return self.participant_id

    def get_index(self, index: str) -> float:
        if index in self.indices:
            return self.indices[index]

        return 0
        
    def get_textual_index(self, index: str) -> float:
        if index in self.textual_complexity_indices:
            return self.textual_complexity_indices[index]
        return 0

    def __eq__(self, other):
        return self.participant_id == other.participant_id
    
    def __hash__(self):
        return hash(self.participant_id)

    def set_index(self, index: str, value: float):
        self.indices[index] = value

    def set_longitudinal_index(self, index: str, value: float):
        self.longitudinal_indices[index] = value

    def set_textual_index(self, index: str, value: float):
        self.textual_complexity_indices[index] = value

    def set_own_conversation(self, conversation):
        self.own_conversation = conversation

    def add_eligible_contribution(self, contribution):
        self.eligible_contributions.append(contribution)

    def set_eligible_contributions(self, contributions):
        self.eligible_contributions = contributions

    def add_significant_contribution(self, contribution):
        self.significant_contributions.append(contribution)

    def export_individual_statistics(self, filename: str):
        with open('mycsvfile.csv','wb') as f:
            w = csv.writer(f)
            w.writerow(somedict.keys())
            w.writerow(somedict.values())



from enum import Enum, unique

@unique
class CNAIndices(Enum):
    CONTRIBUTIONS_SCORE = 'contributions_score'
    SOCIAL_KB = 'social_kb'
    OUTDEGREE = 'outdegree'
    INDEGREE = 'indegree'
    CLOSENESS = 'closeness'
    BETWEENNESS = 'betweenness'
    EIGENVECTOR = 'eigenvector'
    NO_NEW_THREADS = 'no_new_threads'
    NEW_THREADS_OVERALL_SCORE = 'new_threads_overall_score'
    # NEW_THREADS_CUMULATIVE_SOCIAL_KB = 'new_threads_cumulative_social_kb'
    AVERAGE_LENGTH_NEW_THREADS = 'average_length_new_threads'


from rb.core.lang import Lang
from rb.core.text_element import TextElement
from rb.core.word import Word
from rb.complexity.complexity_index import ComplexityIndex
from rb.complexity.index_category import IndexCategory
from rb.core.text_element_type import TextElementType
from rb.complexity.measure_function import MeasureFunction
from typing import Callable, List
from collections import Counter
from rb.utils.rblogger import Logger
import math
from enum import Enum

logger = Logger.get_logger()


class ChNgramEntropyEnum(Enum):
    ONE = 1
    TWO = 2
    THREE = 3


class ChNgramEntropy(ComplexityIndex):
    def __init__(self, lang: Lang, ngram_size: ChNgramEntropyEnum,
        reduce_depth: int, reduce_function: MeasureFunction):

        ComplexityIndex.__init__(self, lang=lang, category=IndexCategory.SURFACE,
                                 abbr="NgramEntr", reduce_depth=reduce_depth,
                                 reduce_function=reduce_function)
        self.ngram_size = ngram_size

    def get_ngrams(self, chars: str) -> List[str]:
        ngrams = []
        for i in range(len(chars) - (self.ngram_size.value - 1)):
            ngrams.append(chars[i:(i + self.ngram_size.value)])
        return ngrams

    def _compute_value(self, element: TextElement) -> float:
        if element.is_word() == True:
            chars = element.text
            ngrams = self.get_ngrams(chars)
            counter = Counter(ngrams)
            n = len(ngrams)
            return sum(- v / n * math.log(v / n) for v in counter.values())
    
    def __repr__(self):
        return f"{self.reduce_function_abbr}({self.abbr}_{self.ngram_size.value} / {self.reduce_depth_abbr})"


from rb.complexity.rhythm.syllabified_dict import SyllabifiedDict
from rb.core.lang import Lang

from typing import List
from queue import Queue

import nltk

# Constants declaration
PHONEMIC_CONSONANTS = {"b", "ch", "d", "dh", "f", "g", "hh", "jh", "k", "l", "m", "n", "ng", "p", "r", "s", "sh", "t", "th", "v", "w", "y", "z", "zh"}
PHONEMIC_VOWELS = {"aa", "ae", "ah", "ao", "aw", "ay", "eh", "er", "ey", "ih", "iy", "ow", "oy", "uh", "uw"}


def is_consonant_sound(sound: str) -> bool:
    return sound in PHONEMIC_CONSONANTS


def is_vocalic_sound(sound: str) -> bool:
    return sound in PHONEMIC_VOWELS


def is_primary_stressed(syllable: List[str]) -> bool:
    """
    Checks if a syllables is primary stressed or not.
    :param syllable: represented as a list of phonemes
    :return: True or False
    """
    return True if any(phoneme.endswith('1') for phoneme in syllable) else False


def get_lang_expansion(lang: Lang) -> str:
    if lang == Lang.EN:
        return "english"
    elif lang == Lang.FR:
        return "french"
    elif lang == Lang.RO:
        return "romanian"
    elif lang == Lang.ES:
        return "spanish"
    elif lang == Lang.DE:
        return "german"
    elif lang == Lang.RU:
        return "russian"
    elif lang == Lang.IT:
        return "italian"
    elif lang == Lang.NL:
        return "dutch"
    else:
        raise Exception('Undefined expansion for abbreviation ' + lang.value)


def get_structures_from_symbols(symbols: List[str]):
    """
    This method is used to extract lists of alliteration or assonance
    from a list of symbols (phonemes) identified by indices.
    :param symbols: list of phonemes
    :return: indices
    """
    structures_indices = list()
    skip_range = 3
    q = Queue()

    q.put(0)
    processed = {i: False for i in range(len(symbols))}
    while not q.empty():
        i = q.get()
        if processed[i]:
            continue
        indices = [i]
        processed[i] = True
        for j in range(i + 1, len(symbols)):
            if symbols[j] == symbols[i]:
                indices.append(j)
                processed[j] = True
            else:
                q.put(j)
                for k in range(j + 1, min(j + skip_range, len(symbols))):
                    if symbols[k] == symbols[i]:
                        i = k
                if i < j:
                    break
        if len(indices) > 1:
            structures_indices.append(indices)

    return structures_indices


def get_rhythmic_structure(lang: Lang, unit: List[str]) -> List[int]:
    """
    This method is implemented according to the definition of rhythmic structure
    formulated by Solomon Marcus in 'Poetica Matematica', 1970
    :param lang: language of the unit
    :param unit: list of words
    :return: a list of integers representing the rhythmic structure of the unit
    """

    syllabified_dict = SyllabifiedDict.get_instance(lang)
    stopwords = nltk.corpus.stopwords.words(get_lang_expansion(lang))

    rhythmic_structure = list()
    cnt = 0
    for word in unit:
        word = word.lower()
        # 1. check if the word is stop word
        if word in stopwords:
            cnt += syllabified_dict[word] if word in syllabified_dict else 1
            continue

        # 2. get the syllables of the word with phonemes (if in dictionary)
        syllables = syllabified_dict[word][0] if word in syllabified_dict else None
        if syllables:
            # 3.1 case when the word was found in dictionary
            for syllable in syllables:
                if is_primary_stressed(syllable):
                    rhythmic_structure.append(cnt)
                    cnt = 0
                else:
                    cnt += 1
        else:
            # 3.2 case when the word was not found in dictionary
            # the word is considered to be unstressed
            # count the number of syllables using a different method
            cnt += 1

    if cnt != 0:
        rhythmic_structure.append(cnt)

    return rhythmic_structure



from statistics import mean
from typing import List, Callable, Dict
import csv
from rb.complexity.complexity_index import ComplexityIndex
from rb.core.lang import Lang
from rb.core.text_element import TextElement
from rb.complexity.index_category import IndexCategory
from rb.complexity.measure_function import MeasureFunction
from rb.complexity.word.aoe_enum import AoeTypeEnum 
from rb.core.text_element_type import TextElementType
from rb.utils.rblogger import Logger

logger = Logger.get_logger()

class Aoe(ComplexityIndex):


    aoe_dict: Dict[AoeTypeEnum, Dict[str, float]] = None

    def __init__(self, lang: Lang, aoe_type: AoeTypeEnum,
            reduce_depth: int, reduce_function: MeasureFunction):

        ComplexityIndex.__init__(self, lang=lang, category=IndexCategory.WORD,
                                 abbr="AoE", reduce_depth=reduce_depth,
                                 reduce_function=reduce_function)
        if Aoe.aoe_dict is None:
            Aoe.parse_aoe_csv(lang)
        self.aoe_type = aoe_type

    @staticmethod
    def parse_aoe_csv(lang: Lang):

        if lang is Lang.EN:
            path_to_aoe = 'resources/en/wordlists/AoE.csv'
        Aoe.aoe_dict = {}
        for at in AoeTypeEnum:
            Aoe.aoe_dict[at] = {}

        with open(path_to_aoe) as csv_file:
            csv_reader = csv.reader(csv_file, delimiter=',')
            for i, row in enumerate(csv_reader):
                if i < 2:
                    continue
                row[0] = row[0].strip()
                if len(str(row[1])) > 0:
                    Aoe.aoe_dict[AoeTypeEnum.INV_AVG][row[0]] = float(row[1])
                if len(str(row[2])) > 0:
                    Aoe.aoe_dict[AoeTypeEnum.INV_LIN_REG_SLOPE][row[0]] = float(row[2])
                if len(str(row[3])) > 0:
                    Aoe.aoe_dict[AoeTypeEnum.IND_ABOVE_THRESH][row[0]] = float(row[3])
                if len(str(row[4])) > 0:
                    Aoe.aoe_dict[AoeTypeEnum.IND_POLY_FIT_ABOVE][row[0]] = float(row[4])
                if len(str(row[5])) > 0:
                    Aoe.aoe_dict[AoeTypeEnum.INFL_POINT_POLY][row[0]] = float(row[5])

    def _compute_value(self, element: TextElement) -> float:
        values = [Aoe.aoe_dict[self.aoe_type][word.text] for word in element.get_words() if word.text in Aoe.aoe_dict[self.aoe_type]]
        return mean(values) if values else 0
    
    def __repr__(self):
        return f"{self.reduce_function_abbr}({self.abbr}_{self.aoe_type.name.lower()} / {self.reduce_depth_abbr})"




from rb.core.lang import Lang
from rb.core.pos import POS
from enum import Enum
from typing import List, Set, Dict
from enum import Enum
from rb.core.pos_features.pos_feature_extractor import POSFeatureExtractor
from conllu import parse, parse_incr
from rb.parser.spacy_parser import computePOS
from rb.core.pos_features.ro_pos_features.ro_features_name import RoFeaturesName
from rb.core.pos_features.ro_pos_features.ro_pos_feature_case import RoPOSFeatureCase, RoCaseEnum
from rb.core.pos_features.ro_pos_features.ro_pos_feature_definite import RoPOSFeatureDefinite, RoDefiniteEnum
from rb.core.pos_features.ro_pos_features.ro_pos_feature_degree import RoPOSFeatureDegree, RoDegreeEnum
from rb.core.pos_features.ro_pos_features.ro_pos_feature_abbr import RoPOSFeatureAbbr, RoAbbrEnum
from rb.core.pos_features.ro_pos_features.ro_pos_feature_adp_type import RoPOSFeatureAdpT, RoAdpTypeEnum
from rb.core.pos_features.ro_pos_features.ro_pos_feature_gender import RoPOSFeatureGender, RoGenderEnum
from rb.core.pos_features.ro_pos_features.ro_pos_feature_mood import RoPOSFeatureMood, RoMoodEnum
from rb.core.pos_features.ro_pos_features.ro_pos_feature_number import RoPOSFeatureNumber, RoNumberEnum
from rb.core.pos_features.ro_pos_features.ro_pos_feature_number_psor import RoPOSFeatureNumberP, RoNumberPEnum
from rb.core.pos_features.ro_pos_features.ro_pos_feature_foreign import RoPOSFeatureForeign, RoForiegnEnum
from rb.core.pos_features.ro_pos_features.ro_pos_feature_num_type import RoPOSFeatureNumType, RoNumTypeEnum
from rb.core.pos_features.ro_pos_features.ro_pos_feature_num_form import RoPOSFeatureNumForm, RoNumFormEnum
from rb.core.pos_features.ro_pos_features.ro_pos_feature_part_type import RoPOSFeaturePartType, RoPartTypeEnum
from rb.core.pos_features.ro_pos_features.ro_pos_feature_person import RoPOSFeaturePerson, RoPersonEnum 
from rb.core.pos_features.ro_pos_features.ro_pos_feature_polarity import RoPOSFeaturePolarity, RoPolarityEnum
from rb.core.pos_features.ro_pos_features.ro_pos_feature_position import RoPOSFeaturePosition, RoPositionEnum
from rb.core.pos_features.ro_pos_features.ro_pos_feature_poss import RoPOSFeaturePoss, RoPossEnum
from rb.core.pos_features.ro_pos_features.ro_pos_feature_pron_type import RoPOSFeaturePronType, RoPronTypeEnum
from rb.core.pos_features.ro_pos_features.ro_pos_feature_strength import RoPOSFeatureStrength, RoStrengthEnum
from rb.core.pos_features.ro_pos_features.ro_pos_feature_tense import RoPOSFeatureTense, RoTenseEnum
from rb.core.pos_features.ro_pos_features.ro_pos_feature_variant import RoPOSFeatureVariant, RoVariantEnum
from rb.core.pos_features.ro_pos_features.ro_pos_feature_verb_form import RoPOSFeatureVerbForm, RoVerbFormEnum

from rb.utils.rblogger import Logger
logger = Logger.get_logger() 


class RoPOSFeatureExtractor(POSFeatureExtractor):

    
    _INSTANCE = None
    UD_RO_FILES = [
        "resources/ro/spacy/ud_tags/ro_rrt-ud-dev.conllu",
        "resources/ro/spacy/ud_tags/ro_rrt-ud-test.conllu",
        "resources/ro/spacy/ud_tags/ro_rrt-ud-train.conllu"]

    def __init__(self):
        self.lang = Lang.RO
        self.create_ud_dict()

    @classmethod
    def get_instance(cls) -> "RoPOSFeatureExtractor":
        if cls._INSTANCE is None:
            cls._INSTANCE = RoPOSFeatureExtractor()
        return cls._INSTANCE

    def create_ud_dict(self) -> Dict[str, Dict[str, List[Dict[Enum, List[Enum]]]]]:
        
        logger.info('Starting parsing ud ro dictionary')
        for ud_file in RoPOSFeatureExtractor.UD_RO_FILES:
            data_file = open(ud_file, "r", encoding="utf-8")
            self.all_forms = {} 

            for tokenlist in parse_incr(data_file):
                for token in tokenlist:
                    s = ""
                    if token['feats'] is not None:
                        for k, v in token['feats'].items():
                            s += '{}={}|'.format(k, v)
                    tag = s
                    pos = computePOS(token['xpostag'], lang=Lang.RO)
                    lemma = token['lemma']
                    form = token['form'].lower()
                    features = self.get_pos_features(pos, tag)
                    if lemma not in self.all_forms:
                        self.all_forms[lemma] = {}
                    if form not in self.all_forms[lemma]:
                        self.all_forms[lemma][form] = []
                    if features not in self.all_forms[lemma][form]:
                        self.all_forms[lemma][form].append(features)
        logger.info('Finished parsing ud ro dictionary')
        return self.all_forms

    def print_ud_dict(self, file_name: str) -> None:
        
        file_handle = open(file_name, 'w', encoding='utf-8')
        all_combinations = set()
        for lemma, forms in self.all_forms.items():
            # print(lemma, len(forms), file=file_handle)
            for word_form, features_sup in forms.items():
                # print('  ', word_form, len(features_sup), file=file_handle)
                for features in features_sup:
                    el = []
                    for key, f in features.items():
                        if len(f) > 0:
                            for ff in f:
                                el.append(ff)
                    print(el, file=file_handle)
                    all_combinations.add(frozenset(el))
                        # if f in features:
                        #     print(features[f], file=file_handle)
        print(len(all_combinations), file=file_handle)

    def get_pos_features(self, pos: POS, tag: str) -> Dict[Enum, List[Enum]]:
        features = {}
        feature_instances = [
            (RoAdpTypeEnum, RoPOSFeatureAdpT.get_instance()),
            (RoCaseEnum, RoPOSFeatureCase.get_instance()),
            (RoAbbrEnum, RoPOSFeatureAbbr.get_instance()),
            (RoDefiniteEnum, RoPOSFeatureDefinite.get_instance()),
            (RoDegreeEnum, RoPOSFeatureDegree.get_instance()),
            (RoGenderEnum, RoPOSFeatureGender.get_instance()),
            (RoMoodEnum, RoPOSFeatureMood.get_instance()),
            (RoNumberEnum, RoPOSFeatureNumber.get_instance()),
            (RoNumberPEnum, RoPOSFeatureNumberP.get_instance()),
            (RoForiegnEnum, RoPOSFeatureForeign.get_instance()),
            (RoNumTypeEnum, RoPOSFeatureNumType.get_instance()),
            (RoNumFormEnum, RoPOSFeatureNumForm.get_instance()),
            (RoPartTypeEnum, RoPOSFeaturePartType.get_instance()),
            (RoPersonEnum, RoPOSFeaturePerson.get_instance()),
            (RoPolarityEnum, RoPOSFeaturePolarity.get_instance()),
            (RoPositionEnum, RoPOSFeaturePosition.get_instance()),
            (RoPossEnum, RoPOSFeaturePoss.get_instance()),
            (RoPronTypeEnum, RoPOSFeaturePronType.get_instance()),
            (RoStrengthEnum, RoPOSFeatureStrength.get_instance()),
            (RoTenseEnum, RoPOSFeatureTense.get_instance()),
            (RoVariantEnum, RoPOSFeatureVariant.get_instance()),
            (RoVerbFormEnum, RoPOSFeatureVerbForm.get_instance())]

        for feature_type, feature_instance in feature_instances:
            #if pos in feature_instance.pos_supported:
            features[feature_type] = feature_instance.get_values(tag)
        return features
    
    def get_all_forms_from_lemma(self, lemma: str) -> Dict[str, Dict[Enum, List[Enum]]]:
        return self.all_forms[lemma]

if __name__ == "__main__":
    feature_extractor = RoPOSFeatureExtractor()
    #feature_extractor.print_ud_dict('log.log')

    for form, feats in feature_extractor.get_all_forms_from_lemma(lemma='vedea').items():
        print(form, feats)


from enum import Enum, unique

@unique
class CsclCriteria(Enum):
    AVERAGE = 'average'
    STDEV = 'stdev'
    SLOPE = 'slope'
    ENTROPY = 'entropy'
    UNIFORMITY = 'uniformity'
    PEAK_CHAT_FRAME = 'peak_chat_frame'
    LOCAL_EXTREME = 'local_extreme'
    RECURRENCE_AVERAGE = 'recurrence_average'
    RECURRENCE_STDEV = 'recurrence_stdev'
    
    


import tensorflow as tf
gpus = tf.config.experimental.list_physical_devices('GPU')
if gpus:
    try:
        for gpu in gpus:
            tf.config.experimental.set_memory_growth(gpu, True)

    except RuntimeError as e:
        print(e)
import tensorflow.keras as keras
import numpy as np
import utils
import sys
import absl
import bert
import functools
import os
from rb.processings.encoders.bert import BertWrapper
from rb.core.lang import Lang


FLAGS = absl.flags.FLAGS
absl.flags.DEFINE_integer('bert_max_seq_len', 128, 'Maximum sequence length')
absl.flags.DEFINE_string('model_folder_path', '../Models/raw_old/full/clean/trained_128_512/ro0/', 'Path to bert model folder')
absl.flags.DEFINE_float('learning_rate', 1e-5, 'Learning Rate used for optimization')
absl.flags.DEFINE_integer('batch_size', 32, 'Batch size to use during training')
absl.flags.DEFINE_integer('epochs', 1, 'Number of epochs to train')
absl.flags.DEFINE_float('dropout_rate', 0.5, 'Dropout rate')
absl.flags.DEFINE_integer('num_classes', 4, "Number of classes for clasification task")
absl.flags.DEFINE_integer('experiment_index', 1, 'Index of current experiment. Will be appended to weights file')
absl.flags.DEFINE_string('save_folder_path',".", "Save folder prefix")
absl.flags.DEFINE_bool("use_tpu", False, "Use TPU or not")
absl.flags.DEFINE_string("tpu_name", None, "Name of TPU instance")

def gen_data(features, labels):
    for index in range(len(features[0])):
        yield ({'input_ids': features[0][index], 'segment_ids': features[1][index]}, labels[index])


def create_model():

    # define input
    input_ids = tf.keras.layers.Input(shape=(FLAGS.bert_max_seq_len), dtype=tf.int32, name="input_ids")
    segment_ids = tf.keras.layers.Input(shape=(FLAGS.bert_max_seq_len), dtype=tf.int32, name="segment_ids")

    # bert_wrapper
    bert_wrapper = BertWrapper(Lang.RO, max_seq_len=FLAGS.bert_max_seq_len, model_name=FLAGS.bert_model_type)
    sys.exit()

    # define model
    bert_model = BertModel.BertModel(FLAGS.model_folder_path, FLAGS.max_seq_len, do_lower_case)
    bert_output = bert_model.bert_layer([input_ids, segment_ids])
    cls_output = keras.layers.Lambda(lambda seq: seq[:, 0, :])(bert_output)
    cls_drop = keras.layers.Dropout(FLAGS.dropout_rate)(cls_output)
    fc1 = keras.layers.Dense(units=100, activation="relu")(cls_drop)
    prediction = keras.layers.Dense(units=FLAGS.num_classes, activation="softmax")(fc1)

    # build model
    model = keras.Model(inputs=[input_ids, segment_ids], outputs=prediction)
    model.build(input_shape=[(None, FLAGS.max_seq_len), (None, FLAGS.max_seq_len)])
    # load pretrained
    bert.load_bert_weights(bert_model.bert_layer, FLAGS.model_folder_path+"bert_model.ckpt")

    model.compile(optimizer=keras.optimizers.Adam(lr=FLAGS.learning_rate), loss = 'categorical_crossentropy', metrics = [tf.keras.metrics.categorical_accuracy])
    model.summary()

    print("Do lower case =", do_lower_case)

    return model, bert_model


def main(argv):
    del argv
        
    model, bert = create_model()


    # load data
    train_data = utils.readJson("../Dataset/Reviews/4Classes/train.json")
    dev_data = utils.readJson("../Dataset/Reviews/4Classes/dev.json")
    # test_data = utils.readJson("../Dataset/Reviews/4Classes/test.json")

    train_features = utils.getFeatures(train_data, sample_majority=True, sample_count=20000, seed=12345)
    train_features, train_labels, train_weights_dict = utils.processFeatures(train_features, bert)
    print(len(train_features[0]), len(train_labels), train_weights_dict)

    dev_features = utils.getFeatures(dev_data)
    dev_features, dev_labels, _ = utils.processFeatures(dev_features, bert)
    print(len(dev_features[0]), len(dev_labels))

    # test_features = utils.getFeatures(test_data)
    # test_features, test_labels, _ = utils.processFeatures(test_features, bert)
    # print(len(test_features[0]), len(test_labels))
    
    t1 = []
    t2 = []
    for i in range(len(train_features[0])):
        t1.append(train_features[0][i])
        t2.append(train_features[1][i])
    train_dataset = tf.data.Dataset.from_tensor_slices(((t1, t2), train_labels))
    train_dataset = train_dataset.shuffle(70000)
    train_dataset = train_dataset.repeat(-1).batch(FLAGS.batch_size)

    d1 = []
    d2 = []
    for i in range(len(dev_features[0])):
        d1.append(dev_features[0][i])
        d2.append(dev_features[1][i])
    dev_dataset = tf.data.Dataset.from_tensor_slices(((d1, d2), dev_labels))
    dev_dataset = dev_dataset.batch(32)

    # dataset for metric :(
    # dev_dataset_metric = tf.data.Dataset.from_generator(functools.partial(gen_data, dev_features, dev_labels), ({'input_ids': tf.int32, 'segment_ids': tf.int32},  tf.int32),
    #         ({'input_ids': tf.TensorShape([FLAGS.max_seq_len]), 'segment_ids': tf.TensorShape([FLAGS.max_seq_len])}, tf.TensorShape([None])))
    # dev_dataset_metric = dev_dataset_metric.batch(FLAGS.batch_size)
    fmetric = utils.FScoreCallback(dev_dataset, len(dev_labels)//32, dev_labels)


    folder_name = FLAGS.model_folder_path.split("/")[-2]+"_"+str(FLAGS.experiment_index)
    os.makedirs(FLAGS.save_folder_path+"/{0}/".format(folder_name))

    results = []
    for i in range(FLAGS.epochs):
        print("EPOCH ", i+1)
        _= model.fit(train_dataset, steps_per_epoch=len(train_labels)//FLAGS.batch_size, epochs=1, verbose=1)    
        model.evaluate(dev_dataset, callbacks=[fmetric])
        model.save(FLAGS.save_folder_path+"/{0}/model{1}.h5".format(folder_name, str(i+1)))


if __name__ == "__main__":

    absl.flags.mark_flag_as_required('model_folder_path')
    absl.flags.mark_flag_as_required('experiment_index')
    absl.app.run(main)


from typing import Callable, List, Tuple, Union
from threading import Lock

from nltk.corpus import wordnet as wn
from nltk.corpus.reader.wordnet import Synset
from rb.core.lang import Lang
from rb.core.word import Word
from rb.parser.spacy_parser import SpacyParser

lang_dict = {
    Lang.EN: 'eng',
    Lang.NL: 'nld',
    Lang.FR: 'fra',
    Lang.RO: 'ron',
    Lang.IT: 'ita'
}

SynsetLock = Lock()

def get_synsets(word: str, lang: str, pos: str = None) -> List[Synset]:
    SynsetLock.acquire()
    result = wn.synsets(word, pos=pos, lang=lang)
    SynsetLock.release()
    return result

def get_synset_hypernyms(ss: Synset) -> List[Synset]:
    SynsetLock.acquire()
    result = ss.hypernyms()
    SynsetLock.release()
    return result

def compute_similarity(a: str, b: str, lang: Lang, sim: Callable[[Synset, Synset], float]) -> float:
    if lang not in lang_dict:
        return 0
    lang = lang_dict[lang]
    return max([
        sim(syn_a, syn_b) or 0
        for syn_a in get_synsets(a, lang=lang)
        for syn_b in get_synsets(b, lang=lang)
        if syn_a.pos() == syn_b.pos()],
        default=0)

def path_similarity(a: str, b: str, lang: Lang) -> float:
    return compute_similarity(a, b, lang, wn.path_similarity)

def leacock_chodorow_similarity(a: str, b: str, lang: Lang) -> float:
    return compute_similarity(a, b, lang, wn.lch_similarity)

def wu_palmer_similarity(a: str, b: str, lang: Lang) -> float:
    return compute_similarity(a, b, lang, wn.wup_similarity)

def get_synonyms(word: Union[str, Word], lang: Lang = None, pos: str = None) -> List[str]:
    if isinstance(word, Word):
        pos = word.pos.to_wordnet()
        lang = word.lang
        word = word.lemma
    if lang not in lang_dict:
        return []
    return list({other
        for ss in get_synsets(word, pos=pos, lang=lang_dict[lang]) 
        for other in ss.lemma_names(lang=lang_dict[lang]) 
        if SpacyParser.get_instance().is_dict_word(other, lang)})

def are_synonyms(first: Union[str, Word], second: Union[str, Word], lang: Lang = None, pos: str = None) -> bool:
    if isinstance(first, Word):
        pos = first.pos.to_wordnet()
        lang = first.lang
        first = first.lemma
    if isinstance(second, Word):
        if second.pos.to_wordnet() !=  pos and pos:
            return False 
        second = second.lemma
    if lang not in lang_dict:
        return False
    first_synset = set(get_synsets(first, pos=pos, lang=lang_dict[lang]))
    second_synset = set(get_synsets(second, pos=pos, lang=lang_dict[lang]))
    return len(first_synset & second_synset) > 0

def are_hypernyms(first: Union[str, Word], second: Union[str, Word], lang: Lang = None, pos: str = None) -> bool:
    if isinstance(first, Word):
        pos = first.pos.to_wordnet()
        lang = first.lang
        first = first.lemma
    if isinstance(second, Word):
        if second.pos.to_wordnet() !=  pos and pos:
            return False 
        second = second.lemma
    if lang not in lang_dict:
        return False
    first_synset = set(get_synsets(first, pos=pos, lang=lang_dict[lang]))
    second_synset = set(get_synsets(second, pos=pos, lang=lang_dict[lang]))
    first_hypernyms = {hypernym for synset in first_synset for hypernym in get_synset_hypernyms(synset)}
    second_hypernyms = {hypernym for synset in second_synset for hypernym in get_synset_hypernyms(synset)}
    return len(first_synset & second_hypernyms) > 0 or len(second_synset & first_hypernyms) > 0

def are_siblings(first: Union[str, Word], second: Union[str, Word], lang: Lang = None, pos: str = None) -> bool:
    if isinstance(first, Word):
        pos = first.pos.to_wordnet()
        lang = first.lang
        first = first.lemma
    if isinstance(second, Word):
        if second.pos.to_wordnet() !=  pos and pos:
            return False 
        second = second.lemma
    if lang not in lang_dict:
        return False
    first_synset = set(get_synsets(first, pos=pos, lang=lang_dict[lang]))
    second_synset = set(get_synsets(second, pos=pos, lang=lang_dict[lang]))
    first_hypernyms = {hypernym for synset in first_synset for hypernym in get_synset_hypernyms(synset)}
    second_hypernyms = {hypernym for synset in second_synset for hypernym in get_synset_hypernyms(synset)}
    return len(first_hypernyms & second_hypernyms) > 0


def get_hypernyms(word: Union[str, Word], lang: Lang = None, pos: str = None) -> List[str]:
    if isinstance(word, Word):
        pos = word.pos.to_wordnet()
        lang = word.lang
        word = word.lemma
    if lang not in lang_dict:
        return []
    return list({other
        for ss in get_synsets(word, pos=pos, lang=lang_dict[lang]) 
        for parent in get_synset_hypernyms(ss)
        for other in parent.lemma_names(lang=lang_dict[lang]) 
        if SpacyParser.get_instance().is_dict_word(other, lang)})

def get_min_path_to_root(ss: Synset, lang: Lang = None, pos: str = None) -> int:
    parents: List[Synset] = get_synset_hypernyms(ss)
    shortest_path = float('Inf')
    for parent in parents:
        path_length = 1 + get_min_path_to_root(parent, lang=lang, pos=pos)
        if  path_length < shortest_path:
            shortest_path = path_length
    
    if shortest_path == float('Inf'):
        return 0
    return shortest_path

def get_all_paths_lengths_to_root(word: Union[str, Word], lang: Lang = None, pos: str = None) -> List[int]:
    if isinstance(word, Word):
        pos = word.pos.to_wordnet()
        lang = word.lang
        word = word.lemma

    if lang not in lang_dict:
        return []
        
    paths = []
    for ss in get_synsets(word, pos=pos, lang=lang_dict[lang]):
        path_length = get_min_path_to_root(ss, lang=lang, pos=pos)
        paths.append(path_length)
    
    return paths
if __name__ == "__main__":
    print(path_similarity('hond', 'kat', 'nl'))
    print(get_hypernyms('animal', Lang.RO))



import pickle

import numpy as np
import rb.processings.diacritics.utils as utils
import tensorflow as tf
from rb.core.lang import Lang
from rb.processings.diacritics.BertCNN import weighted_categorical_crossentropy
from rb.utils.downloader import check_version, download_model
from transformers import AutoTokenizer


class DiacriticsRestoration(object):
    """
    Wrapper for Diacritics restoration
    """

    def __init__(self):

        self.bert_max_seq_len = 512
        self.max_windows = 280
        self.max_sentences = 10
        self.max_sentence_length = 256

        # load model
        if check_version(Lang.RO, ["models", "diacritice", "base"]):
            download_model(Lang.RO, ["models", "diacritice", "base"])

        self.model = tf.keras.models.load_model("resources/ro/models/diacritice/base", compile=False)
        self.tokenizer = AutoTokenizer.from_pretrained("readerbench/RoBERT-base")
        self.char_to_id_dict = pickle.load(open("resources/ro/models/diacritice/base/char_dict", "rb"))


    # replace_all: replaces all diacritics(if existing) with model predictions
    # replace_missing: replaces only characters that accept and don't have diacritics with model predictions; keeps existing diacritics
    def process_string(self, string, mode="replace_all"):
        full_diacritics = set("aăâiîsștț")
        explicit_diacritics = set("ăâîșțĂÂÎȘȚ")
        if len(string) > self.max_sentence_length:
            result = ""
            for i in range(0, len(string), self.max_sentence_length):
                substring = string[i:min(len(string), i+self.max_sentence_length)]
                result += self.process_string(substring, mode)
            return result
        working_string = string.lower()
        clean_string = ""
        # remove everything not in char_to_id_dict
        for s in working_string:
            if s in self.char_to_id_dict.keys():
                clean_string += s

        working_string = clean_string
        working_string = ''.join([utils.get_char_basic(char) for char in working_string])

        diac_count = 0
        for s in working_string:
            if s in full_diacritics:
                diac_count += 1

        x_dataset = tf.data.Dataset.from_generator(lambda : utils.generator_bert_cnn_features_string(working_string, self.char_to_id_dict, 11, self.tokenizer, self.max_sentences, self.max_windows),
                        output_types=({'bert_input_ids': tf.int32, 'bert_segment_ids': tf.int32, 'token_ids': tf.int32, 'sent_ids': tf.int32,
                                        'mask': tf.float32, 'char_windows': tf.int32}, tf.float32),
                        output_shapes=({'bert_input_ids':[self.max_sentences, self.bert_max_seq_len], 'bert_segment_ids':[self.max_sentences, self.bert_max_seq_len], 'token_ids':[self.max_windows],
                                        'sent_ids': [self.max_windows], 'mask': [self.max_windows], 'char_windows': [self.max_windows, 11]}, [self.max_windows, 5]))
        x_dataset = x_dataset.batch(1)

        predictions = self.model.predict(x_dataset, steps=(diac_count//self.max_windows)+1)

        filtered_predictions = []
        for index in range(len(predictions[0])):
            if predictions[1][index] == 1:
                filtered_predictions.append(predictions[0][index])
        
        predictions = np.array(filtered_predictions)
        predicted_classes = list(map(lambda x: np.argmax(x), predictions))
        prediction_index = 0
        
        complete_string = ""
        for orig_char in string:

            lower_orig_char = orig_char.lower()

            if lower_orig_char in full_diacritics:
                if mode == "replace_all":
                    new_char = utils.get_char_from_label(utils.get_char_basic(lower_orig_char), predicted_classes[prediction_index])
                
                elif mode == "replace_missing":					
                    if lower_orig_char in explicit_diacritics:
                        new_char = lower_orig_char
                    else:
                        new_char = utils.get_char_from_label(utils.get_char_basic(lower_orig_char), predicted_classes[prediction_index])
                
                prediction_index += 1

                if orig_char.isupper():
                    new_char = new_char.upper()

            else:
                new_char = orig_char
            complete_string += new_char

        return complete_string



#pylint: disable=import-error

import numpy as np
import tensorflow as tf
from rb.core.lang import Lang
from rb.utils.downloader import check_version, download_model
from transformers import AutoTokenizer, TFAutoModelForSequenceClassification


class SentimentAnalysis(object):
	"""
	Wrapper for Sentiment Analysis
    """

	def __init__(self, lang: Lang, max_seq_len=128, check_updates = True):
		# load model
		self.lang = lang
		self.max_seq_len = min(max_seq_len, 512)
		if self.lang is Lang.RO:
			if check_updates and check_version(Lang.RO, ["models", "sentiment", "base"]):
				download_model(Lang.RO, ["models", "sentiment", "base"])

			self.model = tf.keras.models.load_model("resources/ro/models/sentiment/base")
			self.tokenizer = AutoTokenizer.from_pretrained("readerbench/RoBERT-base")
			self.max_seq_len = 512
		else:
			self.tokenizer = AutoTokenizer.from_pretrained("nlptown/bert-base-multilingual-uncased-sentiment") 
			self.model = TFAutoModelForSequenceClassification.from_pretrained("nlptown/bert-base-multilingual-uncased-sentiment", from_pt=True)
	
		
	def process_text(self, text):

		if isinstance(text, str):
			text = [text]

		if self.lang is Lang.RO:
			features = self.tokenizer(text, return_tensors="tf", padding="max_length", max_length=self.max_seq_len)
			del features['attention_mask']
			predictions = self.model(features).numpy()
			return [x[0] / 5.0 for x in predictions]
		else:
			features = self.tokenizer(text)
			features = {key: np.array(value) for key, value in features.items()}
			predictions = tf.nn.softmax(self.model(features).logits).numpy()
			return [sum(i / 5 * p for i, p in enumerate(pred)) for pred in predictions]
		


from typing import List

from rb.core.lang import Lang
from rb.core.pos import POS
from rb.core.text_element import TextElement
from rb.core.text_element_type import TextElementType
from rb.parser.spacy_parser import SpacyParser
from rb.core.pos_features.pos_feature_extractor import POSFeatureExtractor
from spacy.tokens import Token


class Word(TextElement):

    MODAL_VERBS = {
        Lang.EN: {"can", "may", "must", "shall", "will", "could", "might", "should", "would"},
    }

    def __init__(self, lang: Lang, token: Token, index_in_container: int,
                 depth: int = TextElementType.WORD.value,
                 container: TextElement = None):

        TextElement.__init__(self, lang=lang, text=token.text, index_in_container=index_in_container,
                             depth=depth, container=container)
        self.lemma = token.lemma_
        self.pos = POS(token.pos_)
        self.detailed_pos = token.tag_
        self.head: "Word"
        self.dep: str = token.dep_
        self.ent_type_: str  = token.ent_type_
        self.ent_type = token.ent_type
        self.ent_id_ = token.ent_id_
        self.ent_id = token.ent_id
        self.is_stop = token.is_stop
        self.is_alpha = token.is_alpha
        self.children: List["Word"] = []
        self.tag: str = token.tag_
        self.index_in_doc: int = token.i
        self.in_coref = False
        self.coref_clusters = []
        self.pos_features = None#POSFeatureExtractor.create(lang).get_pos_features(self.pos, self.tag)

    @classmethod
    def from_str(cls, lang: Lang, text: str, pos: POS = POS.X) -> "Word":
        token = type('Token', (object,), {'text': text, 'lemma_': text, 'pos_': pos.value, 'tag_': pos.value, 
                     'dep_': None, 'ent_type_': None, 'ent_type': None, 'ent_id_': None, 'ent_id': None,
                     'is_alpha': None, 'is_stop': None, 'tag': None, 'i': None})()
        return Word(lang, token, None)
    
    def is_dict_word(self) -> bool:
        return SpacyParser.get_instance().is_dict_word(self.lemma, self.lang)

    def is_content_word(self) -> bool:
        return self.is_dict_word() and self.pos in {POS.ADJ, POS.ADV, POS.NOUN, POS.VERB}
    
    def is_modal_verb(self) -> bool:
        if self.lang not in self.MODAL_VERBS:
            return False
        return self.pos is POS.VERB and self.lemma in self.MODAL_VERBS[self.lang]
            
    def get_sentences(self) -> List["Sentence"]:
        return []

    def __str__(self):
        return self.text
    
    def __repr__(self):
        return self.__str__()

    def __hash__(self):
        return hash((self.lemma, self.pos, self.index_in_doc))



from rb.core.span import Span
from rb.core.word import Word
from rb.core.lang import Lang
from rb.core.text_element import TextElement
from neuralcoref.neuralcoref import Cluster
from typing import List, Dict


class CorefCluster():

    def __init__(self, lang: Lang, nc_cluster: Cluster, words: Dict[int, Word]):
        self.lang = lang
        self.index = nc_cluster.i
        self.main = Span.from_spacy_span(lang, nc_cluster.main, words)
        self.mentions = [Span.from_spacy_span(lang, mention, words) for mention in nc_cluster.mentions]
        # update the words
        for mention in nc_cluster.mentions:
            m_words = [words[i] for i in range(mention.start, mention.end)]
            for word in m_words:
                word.in_coref = True
                word.coref_clusters.append(self)



import os
import sys
from os import listdir
from os.path import isdir, isfile, join
from typing import Iterable, List

import spacy
from gensim.models import Word2Vec
from nltk.tokenize import WordPunctTokenizer, sent_tokenize
from rb.core.lang import Lang
from rb.similarity.vector_model_factory import VectorModelType

#from spacy.lang.ro.lex_attrs import words

# nlp = spacy.load('en_core_web_sm')
# nlp.remove_pipe('tagger')
# nlp.remove_pipe('parser')
# nlp.remove_pipe('ner')

class HiddenPrints:
    def __enter__(self):
        self._original_stdout = sys.stdout
        sys.stdout = open(os.devnull, 'w')

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.close()
        sys.stdout = self._original_stdout

def split_sentences(fileName: str) -> Iterable[List[str]]:
    # sentences = []
    tokenizer = WordPunctTokenizer()
    with open(fileName, "rt", encoding='utf-8') as f:
        for line in f.readlines():
            for sent in sent_tokenize(line):
                yield [token for token in tokenizer.tokenize(sent)
                       if token.isalpha and not token == '.']

def tokenize_docs(fileName: str) -> Iterable[List[str]]:
    # sentences = []
    tokenizer = WordPunctTokenizer()
    with open(fileName, "rt", encoding='utf-8') as f:
        for line in f.readlines():
            yield [token for token in tokenizer.tokenize(line)
                if token.isalpha and not token == '.']

def load_docs_all(folder: str) -> List[List[str]]:
    all_docs = []
    for f in listdir(folder):
        if isfile(join(folder, f)) and f.endswith(".txt"):
            with open(join(folder, f), "rt", encoding='utf-8', errors='replace') as fin:
                content = fin.read()
                all_docs.append([f, content])
    return all_docs

def load_docs(folder: str) -> Iterable[str]:
    for f in listdir(folder):
        if isfile(join(folder, f)) and f.endswith(".txt"):
            with open(join(folder, f), "r") as fin:
                current = ""
                line = fin.readline()
                while line != "":
                    if line == "\n":
                        yield current
                        current = ""
                    else:
                        current += line
                    line = fin.readline()
                if current != "":
                    yield current

def str_to_lang(s: str) -> Lang:
    s = s.lower().strip()
    if s in ["ro", "rou", "romanian"]:
        return Lang.RO
    elif s in ["en", "eng", "english"]:
        return Lang.EN
    elif s in ["es", "esp", "spanish"]:
        return Lang.ES
    elif s in ["ru", "rus", "russian"]:
        return Lang.RU
    elif s in ["de", "deu", "ger", "german", "deutsch"]:
        return Lang.DE
    elif s in ["fr", "fra", "fre", "french"]:
        return Lang.FR
    elif s in ["it", "ita", "italian"]:
        return Lang.IT
    elif s in ["nl", "nld", "dut", "dutch"]:
        return Lang.NL
    return None

def str_to_vmodel(s: str) -> VectorModelType:
    s = s.lower()
    if s.strip() == "lda":
        return VectorModelType.LDA
    elif s.strip() == "lsa":
        return VectorModelType.LSA
    elif s.strip() == "word2vec":
        return VectorModelType.WORD2VEC
    return VectorModelType.WORD2VEC

if __name__ == "__main__":
    pass
    # with open("dict_ro.txt", "wt") as out:
    #     all_words = words | {word for word in LOOKUP} | {lemma for word, lemma in LOOKUP.items()}
    #     all_words = sorted(list(all_words))
    #     for word in all_words:
    #         out.write(word + "\n")



from rb.comprehension.utils.distance_strategies.cm_syntactic_graph import CmSyntacticGraph
from rb.comprehension.utils.graph.cm_edge_do import CmEdgeDO
from rb.comprehension.utils.graph.cm_edge_type import CmEdgeType
from rb.comprehension.utils.distance_strategies.abstract_distance_strategy import DistanceStrategy


from rb.core.word import Word


class SyntacticWordDistanceStrategy(DistanceStrategy):

    def __init__(self, syntactic_graph: CmSyntacticGraph):
        self.edge_list = syntactic_graph.edge_list

    
    def get_distance(self, first_word: Word, second_word: Word) -> float:
        for edge in self.edge_list:
            dependent_word = edge.node1.word
            governor_word = edge.node2.word

            if (dependent_word == first_word and governor_word == second_word) or \
                (dependent_word == second_word and governor_word == first_word):
                return 1.0
        
        return 0.0


    def get_edge_type(self) -> CmEdgeType:
        return CmEdgeType.Syntactic


from datetime import datetime
from typing import List, Dict

from rb.core.block import Block
from rb.core.lang import Lang
from rb.core.sentence import Sentence
from rb.core.text_element import TextElement
from rb.core.block import Block
from rb.core.text_element_type import TextElementType
from rb.core.cscl.participant import Participant
from rb.parser.spacy_parser import SpacyParser
from rb.utils.rblogger import Logger

logger = Logger.get_logger()

SIGNIFICANT_LIMIT = 5


class Contribution(TextElement):

    def __init__(self, lang: Lang, text: str, participant: Participant, index_in_container: int,
                 parent_contribution: "Contribution", contribution_raw: Dict, index: int,
                 timestamp: datetime, depth: int = TextElementType.BLOCK.value,
                 container: TextElement = None):
        super().__init__(lang=lang, text=text, depth=depth, container=container, index_in_container=index_in_container)
        self.parent_contribution = parent_contribution
        self.participant = participant
        self.timestamp = timestamp
        self.index = index

        # used for creating per-participant conversations, based on original contributions
        self.contribution_raw = contribution_raw

    def add_sentence(self, sentence: Sentence):
        self.components.append(sentence)
        sentence.container = self

    def get_parent(self) -> "Contribution":
        return self.parent_contribution

    def get_participant(self) -> Participant:
        return self.participant

    def get_timestamp(self) -> int:
        return self.timestamp

    def is_significant(self):
        return len(self.get_words()) >= SIGNIFICANT_LIMIT

    def get_raw_contribution(self) -> "Contribution":
        return self.contribution_raw
    
    def __eq__(self, other):
        if not isinstance(other, Contribution):
            return False
        return self.timestamp == other.timestamp and TextElement.__eq__(self, other)
    
    def __hash__(self):
        return hash((self.timestamp, TextElement.__hash__(self)))
    


from typing import Dict, List

import numpy as np
from rb.complexity.complexity_index import ComplexityIndex
from rb.core.document import Document
from rb.processings.pipeline.dataset import Dataset, Task
from sklearn.base import BaseEstimator
from sklearn.model_selection import cross_val_score
from sklearn.metrics import accuracy_score, make_scorer, mean_squared_error


class Estimator:
    def __init__(self, dataset: Dataset, tasks: List[Task], params: Dict[str, str]):
        self.dataset = dataset
        self.tasks = tasks
        self.model: BaseEstimator = None
        self.scoring = "accuracy"

    def predict(self, indices: Dict[ComplexityIndex, float]):
        return self.model.predict(self.construct_input(indices))

    def construct_input(self, indices: Dict[ComplexityIndex, float]) -> np.ndarray:
        result = [indices[feature]
                  for i, feature in enumerate(self.dataset.features)
                  if self.tasks[0].mask[i]]
        return np.array(result)

    def cross_validation(self, n=5) -> float:
        x = [self.construct_input(indices) for indices in self.dataset.normalized_train_features]
        y = self.tasks[0].get_train_targets()
        scores = cross_val_score(self.model, x, y, scoring=make_scorer(self.scoring), cv=n)
        return scores.mean()
    
    def evaluate(self) -> float:
        x = [self.construct_input(indices) for indices in self.dataset.normalized_train_features]
        y = self.tasks[0].get_train_targets()
        self.model.fit(x, y)
        x = [self.construct_input(indices) for indices in self.dataset.normalized_dev_features]
        y = self.tasks[0].get_dev_targets()
        predicted = self.model.predict(x)
        return self.scoring(y, predicted)

    @classmethod
    def parameters(cls) -> Dict[str, List]:
        return {}
    
    @classmethod
    def valid_config(cls, config) -> bool:
        return True

class Classifier(Estimator):
    def __init__(self, dataset: Dataset, tasks: List[Task], params: Dict[str, str]):
        super().__init__(dataset, tasks, params)
        self.scoring = accuracy_score

class Regressor(Estimator):
    def __init__(self, dataset: Dataset, tasks: List[Task], params: Dict[str, str]):
        super().__init__(dataset, tasks, params)
        self.scoring = mean_squared_error




from rb.complexity.complexity_index import ComplexityIndex
from rb.core.lang import Lang
from rb.core.text_element import TextElement
from rb.core.word import Word
from rb.complexity.index_category import IndexCategory
from rb.complexity.measure_function import MeasureFunction
from rb.core.text_element_type import TextElementType   
from typing import List, Callable
from rb.utils.rblogger import Logger

logger = Logger.get_logger()


class ParseDepTree(ComplexityIndex):

    
    def __init__(self, lang: Lang, reduce_depth: int,
            reduce_function: MeasureFunction):

        ComplexityIndex.__init__(self, lang=lang, category=IndexCategory.SYNTAX,
                                 abbr="ParseDepth", reduce_depth=reduce_depth,
                                 reduce_function=reduce_function)

    def _compute_value(self, element: TextElement) -> int:
        return self.compute_height(element.root)

    def compute_height(self, word: Word) -> int:
        if not list(word.children):
            return 1
        else:
            return 1 + max(self.compute_height(x) for x in word.children)


from rb.core.lang import Lang
from rb.core.text_element import TextElement
from rb.complexity.complexity_index import ComplexityIndex
from rb.complexity.index_category import IndexCategory
from rb.complexity.measure_function import MeasureFunction


class RhythmicCoefficient(ComplexityIndex):

    def __init__(self, lang: Lang, reduce_depth: int,
                 reduce_function: MeasureFunction):
        ComplexityIndex.__init__(self, lang=lang, category=IndexCategory.RHYTHM,
                                 abbr="RhCoeff", reduce_depth=reduce_depth,
                                 reduce_function=reduce_function)

    def process(self, element: TextElement) -> float:
        pass



import tensorflow as tf
from rb.processings.gec.transformer.utils import positional_encoding
from rb.processings.gec.transformer.encoder_layer import EncoderLayer

class Encoder(tf.keras.layers.Layer):
    def __init__(self, num_layers, d_model, num_heads, dff, input_vocab_size,
               maximum_position_encoding, rate=0.1):
        super(Encoder, self).__init__()

        self.d_model = d_model
        self.num_layers = num_layers

        self.embedding = tf.keras.layers.Embedding(input_vocab_size, d_model)
        self.pos_encoding = positional_encoding(maximum_position_encoding, 
                                                self.d_model)


        self.enc_layers = [EncoderLayer(d_model, num_heads, dff, rate) 
                            for _ in range(num_layers)]

        self.dropout = tf.keras.layers.Dropout(rate)
            
    def call(self, x, training, mask):
        seq_len = tf.shape(x)[1]

        # adding embedding and position encoding.
        x = self.embedding(x)  # (batch_size, input_seq_len, d_model)
        x *= tf.math.sqrt(tf.cast(self.d_model, tf.float32))
        x += self.pos_encoding[:, :seq_len, :]

        x = self.dropout(x, training=training)

        for i in range(self.num_layers):
            x = self.enc_layers[i](x, training, mask)
        return x  # (batch_size, input_seq_len, d_model)


import platform
import sys
from os import chdir, getcwd, makedirs, rmdir
from shutil import rmtree
from subprocess import check_call
from tempfile import TemporaryDirectory

import setuptools
from setuptools.command.develop import develop
from setuptools.command.install import install

with open('requirements.txt') as f:
    required = f.read().splitlines()

def do_post_install_tasks():
    check_call([sys.executable, "-m", "pip", "install", "-r", "requirements.txt"])

    if platform.system() == "Darwin" and platform.processor() == "arm":
        check_call([sys.executable, "-m", "pip", "install", 
            "--pre", "torch", "-f", 
            "https://download.pytorch.org/whl/nightly/cpu/torch_nightly.html"])
        check_call([sys.executable, "-m", "pip", "install", "tensorflow-macos"]) 
        check_call([sys.executable, "-m", "pip", "install", "tensorflow-metal"]) 
        check_call(["brew", "install", "rust"]) 
        cwd = getcwd()
        try:
            import tokenizers
        except:
            with TemporaryDirectory() as temp_folder:
                chdir(temp_folder)
                check_call(["git", "clone", "https://github.com/huggingface/tokenizers"])
                chdir("tokenizers/bindings/python")
                check_call([sys.executable, "-m", "pip", "install", "setuptools_rust"])
                check_call([sys.executable, "setup.py", "install"])      
                chdir(cwd)
        try:
            import transformers
        except:
            check_call([sys.executable, "-m", "pip", "install", "git+https://github.com/huggingface/transformers"])
    else:
        check_call([sys.executable, "-m", "pip", "install", "tensorflow", "torch", "transformers"])

    #  download nltk stuff
    import nltk

    nltk.download('punkt')
    nltk.download('wordnet')
    nltk.download('omw-1.4')

class PostDevelopmentCommand(develop):
    """Post-instalation for development? mode."""
    def run(self):
        develop.run(self)
        do_post_install_tasks()

class PostInstallCommand(install):
    """Post-instalation for install mode."""
    def run(self):
        install.run(self)
        do_post_install_tasks()
        
with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='rbpy-rb',
    version='0.11.8',
    python_requires='>=3.6,<3.10',
    author='Woodcarver',
    author_email='batpepastrama@gmail.com',
    description='ReaderBench library written in python',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/readerbench/ReaderBench',
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=required,
    cmdclass={
        'develop': PostDevelopmentCommand,
        'install': PostInstallCommand,
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent"
    ],
)



from rb.core.lang import Lang
from rb.core.document import Document

from rb.docs_processing.author import Author

from typing import List

import datetime


class Article:

    def __init__(self, title: str, abstract: str, source: str, authors: List[str], date: str, lang: Lang, additional_info: dict):
        self.title = title
        self.abstract = abstract

        self.author_list = []
        self.extract_authors(authors)
        
        self.source = source
        self.date = datetime.datetime.strptime(date, "%d-%m-%Y") if date else None
        self.document = Document(lang, abstract)
        
        self.additional_info = additional_info

    def extract_authors(self, authors):
        for author_name in authors:
            if not author_name:
                continue
            self.author_list.append(author_name.strip())

    def add_this_article_to_its_authors(self, graph):
        for author_name in self.author_list:
            self.add_this_article_to_author(author_name, graph)

    def add_this_article_to_author(self, author_name, graph):
        if author_name not in graph.authors_dict:
            author = Author(author_name)
            author.articles.add(self)
            graph.authors_dict[author_name] = author
        else:
            graph.authors_dict[author_name].articles.add(self)

    def add_this_article_to_the_graph(self, graph):
        graph.articles_set.add(self)

    def __eq__(self, other):
        if isinstance(other, Article):
            return self.title == other.title
        return NotImplemented

    def __hash__(self):
        return hash(tuple(self.title))


def create_article_and_add_it_to_its_authors_and_graph(title: str, abstract: str, source: str,
                                                       authors: List[str], date: str, lang: Lang, graph):
    article = Article(title, abstract, source, authors, date, lang)
    article.add_this_article_to_its_authors(graph)
    article.add_this_article_to_the_graph(graph)

    return article


from rb.core.word import Word
from rb.comprehension.utils.graph.cm_edge_type import CmEdgeType

class DistanceStrategy():

    def get_distance(self, first_word: Word, second_word: Word) -> float:
        pass

    def get_edge_type(self) -> CmEdgeType:
        pass


import os
from enum import Enum
from rb.similarity.vector_model import VectorModel, VectorModelType
from typing import Dict, List, Tuple, Union

import numpy as np
from numpy.linalg import norm
from rb.core.lang import Lang
from rb.core.text_element import TextElement
from rb.core.word import Word
from rb.similarity.vector import Vector
from rb.utils.downloader import check_version, download_model
from rb.utils.rblogger import Logger

WordSimilarity = Tuple[str, float]

logger = Logger.get_logger()

class CorporaEnum(Enum):
    README = 'readme'
    COCA = 'coca'
    JOSE_ANTONIO = 'jose_antonio'
    RNC_WIKIPEDIA = 'rnc_wikipedia'
    LE_MONDE = 'le_monde'
    WIKI = 'wiki'


class WordVectorModel(VectorModel):
    
    def __init__(self, type: VectorModelType, corpus: str, lang: Lang, size: int = 300, check_updates=True):
        VectorModel.__init__(self, type=type, name=f"{type.name}({corpus})", lang=lang, size=size)
        self.corpus = corpus
        self.vectors: Dict[str, Vector] = {}
        self.base_vectors: List[Vector] = []
        self.word_clusters: Dict[int, List[str]] = {}
        if check_updates and check_version(lang, self.corpus):
            if not download_model(lang, self.corpus):
                raise FileNotFoundError("Requested model ({}) not found for {}".format(self.corpus, lang.value))
        logger.info('Loading vectors.. ')
        self.load_vectors()
        logger.info('Vectors loaded')
        if len(self.vectors) > 100000:
            try:
                logger.info('Loading clusters.. ')
                self.load_clusters()
                logger.info('Clusters loaded')
            except:
                logger.info('Clusters not found! Building clusters.. ')
                self.build_clusters(8)
                self.save_clusters()
                logger.info('Clusters saved')
    
    def load_vectors(self):
        self.load_vectors_from_txt_file("resources/{}/models/{}/{}.model".format(self.lang.value, self.corpus, self.type.name))

    def get_vector(self, elem: Union[str, TextElement]) -> Vector:
        if isinstance(elem, str):
            return self.vectors[elem]
        if not self in elem.vectors:
            if isinstance(elem, Word):
                if elem.text in self.vectors:
                    elem.vectors[self] = self.vectors[elem.text]
                elif elem.lemma in self.vectors:
                    elem.vectors[self] = self.vectors[elem.lemma]
                else:
                    return None
            else:
                vectors = [self.get_vector(child) for child in elem.components]
                vectors = [v.values for v in vectors if v is not None]
                elem.vectors[self] = Vector(np.sum(vectors, axis=0)) if len(vectors) > 0 else None
        return elem.vectors[self]
    
    def similarity(self, a: Union[TextElement, Vector], b: Union[TextElement, Vector]) -> float:
        if isinstance(a, TextElement) and isinstance(b, TextElement) and a == b:
            return 1.0
        if isinstance(a, TextElement):
            a = self.get_vector(a)
        if isinstance(b, TextElement):
            b = self.get_vector(b)
        if a is None or b is None:
            return 0.0
        return Vector.cosine_similarity(a, b)

    """
    Loads vectors from file.
    If this method is used the self.size will be overwriten.
    """
    def load_vectors_from_txt_file(self, filepath: str) -> None:
        with open(filepath, 'r', encoding='utf-8') as f:
            line = f.readline()
            line_split = line.split()
            no_of_words = int(line_split[0])
            no_of_dimensions = int(line_split[1])
            self.size = no_of_dimensions

            for line in f:
                line_split = line.split()
                word = line_split[0]
                self.vectors[word] = Vector(np.array(line_split[1:], dtype=np.float))
            
    def compute_hash(self, v: Vector) -> int:
        result = 0
        for base in self.base_vectors:
            # print('base={}'.format(base))
            # print('values={}'.format(v.values))
            # print('result={}'.format(result))
            # print('base.shape={},len(v.values)={}'.format(len(base), len(v.values)))
            # print('np.dot={}'.format(np.dot(base, v.values)))
            # print('int={}'.format(int(np.dot(base, v.values) >= 0)))
            # print('result={}'.format((result << 1) + int(np.dot(base, v.values) >= 0)))
            result = (result << 1) + int(np.dot(base.values, v.values) >= 0)
        return result

    def build_clusters(self, n: int = 12):
        self.base_vectors = [Vector(v) for v in np.random.normal(size=(n, self.size))]
        self.word_clusters = {}
        for w, v in self.vectors.items():
            hash = self.compute_hash(v)
            if hash not in self.word_clusters:
                self.word_clusters[hash] = []
            self.word_clusters[hash].append(w) 

    def save_clusters(self):
        folder = f"resources/{self.lang.value}/models/{self.corpus}"
        os.makedirs(folder, exist_ok=True)     
    
        with open(f"{folder}/{self.type.name}-clusters.txt", "wt", encoding='utf-8') as f:
            f.write("{}\n".format(len(self.base_vectors)))
            for base in self.base_vectors:
                f.write(" ".join(str(x) for x in base.values) + "\n")
            f.write("{}\n".format(len(self.vectors)))
            for hash, words in self.word_clusters.items():
                for word in words:
                    f.write("{} {}\n".format(word, hash))

    def load_clusters(self):
        with open(f"resources/{self.lang.value}/models/{self.corpus}/{self.type.name}-clusters.txt", "rt", encoding="utf-8") as f:
            n = int(f.readline())
            for i in range(n):
                line = f.readline()
                self.base_vectors.append(Vector(np.array([float(x) for x in line.split(" ")])))
            n = int(f.readline())
            for i in range(n):
                line = f.readline()
                word, hash = line.split(" ")
                hash = int(hash)
                if hash not in self.word_clusters:
                    self.word_clusters[hash] = []
                self.word_clusters[hash].append(word) 

    def get_cluster(self, hash: int, vector: Vector, threshold: float = None) -> List[WordSimilarity]:
        if len(self.base_vectors) == 0:
            result = self.vectors.keys()
        else:
            result = self.word_clusters[hash] if hash in self.word_clusters else []
        result = [(word, self.similarity(self.get_vector(word), vector)) for word in result]
        if threshold is None:
            return result
        else:
            return [(word, sim) for word, sim in result if sim > threshold]

    def most_similar(self, elem: Union[str, TextElement, Vector], 
                    topN: int = 10, threshold: float = None) -> List[WordSimilarity]:
        if not isinstance(elem, type(Vector)):
            elem = self.get_vector(elem)
        if elem is None:
            return []
        hash = self.compute_hash(elem)
        cluster = self.get_cluster(hash, elem, threshold)
        if len(cluster) < topN:
            for i in range(len(self.base_vectors) - 1):
                new_hash = hash ^ (1 << (len(self.base_vectors) - i - 1))
                for j in range(i+1, len(self.base_vectors)):
                    new_hash = new_hash ^ (1 << (len(self.base_vectors) - j - 1))
                    cluster = cluster + self.get_cluster(new_hash, elem, threshold)
        return sorted(cluster, key=lambda x: x[1], reverse=True)[:topN]





import os
from rb.core.word import Word
from typing import Dict, List, Union

import numpy as np
import tensorflow as tf
from rb import Block, Document
from rb.core.lang import Lang
from rb.core.text_element import TextElement
from rb.similarity.vector import Vector
from rb.similarity.vector_model import VectorModel, VectorModelType
from rb.utils.rblogger import Logger
from transformers import AutoConfig, AutoTokenizer, TFAutoModel

logger = Logger.get_logger()


class TransformersEncoder(VectorModel):

    def __init__(self, lang: Lang, name: str = None, max_seq_len=512):
        from_pt = False
        if name is None:
            if lang is Lang.EN:
                # name = "roberta-base"
                name = "sentence-transformers/all-distilroberta-v1"
                from_pt=True
                self.SOW = "Ġ"
            elif lang is Lang.FR:
                name = "camembert-base"
                self.SOW = "▁"
            elif lang is Lang.RO:
                name = "readerbench/RoBERT-base"
                self.SOW = "##"
            elif lang is Lang.RU:
                name = "DeepPavlov/rubert-base-cased"
                from_pt=True
                self.SOW = "##"
        VectorModel.__init__(self, VectorModelType.TRANSFORMER, name=name, lang=lang, size=768)  
        config = AutoConfig.from_pretrained(self.name)
        config.output_hidden_states = True
        self.tokenizer = AutoTokenizer.from_pretrained(self.name)
        self.bert = TFAutoModel.from_pretrained(self.name, config=config, from_pt=from_pt)
        self.max_seq_len = max_seq_len
        self.bos = self.tokenizer.bos_token_id if self.tokenizer.bos_token_id is not None else self.tokenizer.cls_token_id
        self.eos = self.tokenizer.eos_token_id if self.tokenizer.eos_token_id is not None else self.tokenizer.sep_token_id

    @staticmethod
    def common_prefix(word1: str, word2: str) -> str:
        i = 0
        while i < min(len(word1), len(word2)):
            if word1[i] != word2[i]:
                return word1[:i]
            i += 1
        return word1[:i]

    def clean_text(self, text: str) -> str:
        return text.replace("’", "'") \
            .replace("“", '"') \
            .replace("”", '"')

    def _create_word_token_dict(self, blocks: List[Block], tokenized: List[str]) -> Dict[Word, List[int]]:
        i = 0
        result = {}
        words = [self.clean_text(word.text) for block in blocks for word in block.get_words()]
        if getattr(self.tokenizer, "do_lower_case", False):
            words = [word.lower() for word in words]
        block_symbols = {s for word in words for s in word}
        tokens = ["".join(s for s in token if s in block_symbols) for token in tokenized]
        current = ""
        for text, word in zip(words, [word for block in blocks for word in block.get_words()]):
            while not tokens[i]:
                i += 1
            current = self.common_prefix(text, tokens[i])
            if not current:
                continue 
            ids = [i]
            if text == current:
                if len(text) < len(tokens[i]):
                    tokens[i] = tokens[i][len(current):]
                else:
                    i += 1
            else: 
                i += 1
                while i < len(tokens) and text.startswith(current + tokens[i]):
                    ids.append(i)
                    current += tokens[i]
                    i += 1
                current = text[len(current):]
                if len(current) > 0:
                    tokens[i] = tokens[i][len(current):]
            result[word] = ids
        return result

    def _encode_block(self, blocks: List[Block]):
        text = "\n".join(block.text for block in blocks)
        tokenized = self.tokenizer.tokenize(self.clean_text(text))
        if not tokenized:
            return
        token_index = self._create_word_token_dict(blocks, tokenized)
        start = 0
        n = len(tokenized)
        token_ids = self.tokenizer.convert_tokens_to_ids(tokenized)
        embeddings = []
        while start < n:
            end = min(n, start + self.max_seq_len - 2)
            input_ids = tf.constant([[self.bos] + token_ids[start:end] + [self.eos]])
            attention_mask = tf.ones(input_ids.shape, dtype=np.int32)
            outputs = self.bert(input_ids=input_ids, attention_mask=attention_mask)
            embeddings.append(outputs.last_hidden_state[0, 1:-1, :].numpy())
            start = end
        embeddings = np.concatenate(embeddings, axis=0)
        for block in blocks:
            for sentence in block.get_sentences():
                for word in sentence.get_words():
                    if word in token_index:
                        word.vectors[self] = Vector(np.mean(embeddings[token_index[word]], axis=0, dtype=np.float64))
                sentence_ids = [
                    i
                    for word in sentence.get_words() 
                    if word in token_index 
                    for i in token_index[word]
                ]
                if sentence_ids:
                    sentence.vectors[self] = Vector(np.mean(embeddings[sentence_ids], axis=0, dtype=np.float64))
            block_ids = [
                i
                for word in block.get_words() 
                if word in token_index 
                for i in token_index[word]
            ]
            if block_ids:
                block.vectors[self] = Vector(np.mean(embeddings[block_ids], axis=0, dtype=np.float64))
        
    def encode(self, document: Document):
        vectors = []
        buffer = []
        for block in document.components:
            if sum(len(x.text) for x in buffer) + len(block.text) < 1000:
                buffer.append(block)
                continue
            try:
                self._encode_block(buffer)
            except Exception as e:
                text = '\n'.join(x.text for x in buffer)
                logger.warning("Invalid characters:\n" + text)
            for x in buffer:
                if self in x.vectors:
                    vectors.append(x.vectors[self].values)
            buffer = [block]
        try:
            self._encode_block(buffer)
        except Exception as e:
            text = '\n'.join(x.text for x in buffer)
            logger.warning("Invalid characters:\n" + text)
        for x in buffer:
            if self in x.vectors:
                vectors.append(x.vectors[self].values)
        if vectors:
            document.vectors[self] = Vector(np.mean(vectors, axis=0))
            
    def most_similar(self, elem: Union[str, TextElement, Vector], 
                    topN: int = 10, threshold: float = None):
        return []